# Copyright 2013 Rackspace Hosting Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import netaddr
from neutron.common import config as neutron_cfg
from neutron import quota
from neutron_lib import exceptions as n_exc
from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import importutils

from quark import allocation_pool
from quark.db import api as db_api
from quark import exceptions as q_exc
from quark import network_strategy
from quark.plugin_modules import ip_policies
from quark.plugin_modules import routes
from quark import plugin_views as v
from quark import quota_driver as qdv
from quark import utils

CONF = cfg.CONF
LOG = logging.getLogger(__name__)
STRATEGY = network_strategy.STRATEGY

quark_subnet_opts = [
    cfg.BoolOpt('allow_allocation_pool_update',
                default=False,
                help=_('Controls whether or not to allow allocation_pool '
                       'updates')),
    cfg.BoolOpt('allow_allocation_pool_growth',
                default=False,
                help=_('Controls whether or not to allow allocation_pool '
                       'growing. Otherwise shrinking is only allowed'))
]

CONF.register_opts(quark_subnet_opts, "QUARK")

ipam_driver = (importutils.import_class(CONF.QUARK.ipam_driver))()


def _validate_subnet_cidr(context, network_id, new_subnet_cidr):
    """Validate the CIDR for a subnet.

    Verifies the specified CIDR does not overlap with the ones defined
    for the other subnets specified for this network, or with any other
    CIDR if overlapping IPs are disabled.

    """
    if neutron_cfg.cfg.CONF.allow_overlapping_ips:
        return

    try:
        new_subnet_ipset = netaddr.IPSet([new_subnet_cidr])
    except TypeError:
        LOG.exception("Invalid or missing cidr: %s" % new_subnet_cidr)
        raise n_exc.BadRequest(resource="subnet",
                               msg="Invalid or missing cidr")

    # Using admin context here, in case we actually share networks later
    subnet_list = db_api.subnet_find(context.elevated(), None, None, None,
                                     False, network_id=network_id,
                                     shared=[False])
    for subnet in subnet_list:
        if (netaddr.IPSet([subnet.cidr]) & new_subnet_ipset):
            # don't give out details of the overlapping subnet
            err_msg = (_("Requested subnet with cidr: %(cidr)s for "
                         "network: %(network_id)s overlaps with another "
                         "subnet") %
                       {'cidr': new_subnet_cidr,
                        'network_id': network_id})
            LOG.error(_("Validation for CIDR: %(new_cidr)s failed - "
                        "overlaps with subnet %(subnet_id)s "
                        "(CIDR: %(cidr)s)"),
                      {'new_cidr': new_subnet_cidr,
                       'subnet_id': subnet.id,
                       'cidr': subnet.cidr})
            raise n_exc.InvalidInput(error_message=err_msg)


def create_subnet(context, subnet):
    """Create a subnet.

    Create a subnet which represents a range of IP addresses
    that can be allocated to devices

    : param context: neutron api request context
    : param subnet: dictionary describing the subnet, with keys
        as listed in the RESOURCE_ATTRIBUTE_MAP object in
        neutron/api/v2/attributes.py.  All keys will be populated.
    """
    LOG.info("create_subnet for tenant %s" % context.tenant_id)
    net_id = subnet["subnet"]["network_id"]

    with context.session.begin():
        net = db_api.network_find(context, None, None, None, False,
                                  id=net_id, scope=db_api.ONE)
        if not net:
            raise n_exc.NetworkNotFound(net_id=net_id)

        sub_attrs = subnet["subnet"]

        always_pop = ["enable_dhcp", "ip_version", "first_ip", "last_ip",
                      "_cidr"]
        admin_only = ["segment_id", "do_not_use", "created_at",
                      "next_auto_assign_ip"]
        utils.filter_body(context, sub_attrs, admin_only, always_pop)

        _validate_subnet_cidr(context, net_id, sub_attrs["cidr"])

        cidr = netaddr.IPNetwork(sub_attrs["cidr"])

        err_vals = {'cidr': sub_attrs["cidr"], 'network_id': net_id}
        err = _("Requested subnet with cidr: %(cidr)s for "
                "network: %(network_id)s. Prefix is too small, must be a "
                "larger subnet. A prefix less than /%(prefix)s is required.")

        if cidr.version == 6 and cidr.prefixlen > 64:
            err_vals["prefix"] = 65
            err_msg = err % err_vals
            raise n_exc.InvalidInput(error_message=err_msg)
        elif cidr.version == 4 and cidr.prefixlen > 30:
            err_vals["prefix"] = 31
            err_msg = err % err_vals
            raise n_exc.InvalidInput(error_message=err_msg)
        # Enforce subnet quotas
        net_subnets = get_subnets(context,
                                  filters=dict(network_id=net_id))
        if not context.is_admin:
            v4_count, v6_count = 0, 0
            for subnet in net_subnets:
                if netaddr.IPNetwork(subnet['cidr']).version == 6:
                    v6_count += 1
                else:
                    v4_count += 1

            if cidr.version == 6:
                tenant_quota_v6 = context.session.query(qdv.Quota).filter_by(
                    tenant_id=context.tenant_id,
                    resource='v6_subnets_per_network').first()
                if tenant_quota_v6 != -1:
                    quota.QUOTAS.limit_check(
                        context, context.tenant_id,
                        v6_subnets_per_network=v6_count + 1)
            else:
                tenant_quota_v4 = context.session.query(qdv.Quota).filter_by(
                    tenant_id=context.tenant_id,
                    resource='v4_subnets_per_network').first()
                if tenant_quota_v4 != -1:
                    quota.QUOTAS.limit_check(
                        context, context.tenant_id,
                        v4_subnets_per_network=v4_count + 1)

        # See RM981. The default behavior of setting a gateway unless
        # explicitly asked to not is no longer desirable.
        gateway_ip = utils.pop_param(sub_attrs, "gateway_ip")
        dns_ips = utils.pop_param(sub_attrs, "dns_nameservers", [])
        host_routes = utils.pop_param(sub_attrs, "host_routes", [])
        allocation_pools = utils.pop_param(sub_attrs, "allocation_pools", None)

        sub_attrs["network"] = net
        new_subnet = db_api.subnet_create(context, **sub_attrs)

        cidrs = []
        alloc_pools = allocation_pool.AllocationPools(sub_attrs["cidr"],
                                                      allocation_pools)
        if isinstance(allocation_pools, list):
            cidrs = alloc_pools.get_policy_cidrs()

        quota.QUOTAS.limit_check(
            context,
            context.tenant_id,
            alloc_pools_per_subnet=len(alloc_pools))

        ip_policies.ensure_default_policy(cidrs, [new_subnet])
        new_subnet["ip_policy"] = db_api.ip_policy_create(context,
                                                          exclude=cidrs)

        quota.QUOTAS.limit_check(context, context.tenant_id,
                                 routes_per_subnet=len(host_routes))

        default_route = None
        for route in host_routes:
            netaddr_route = netaddr.IPNetwork(route["destination"])
            if netaddr_route.value == routes.DEFAULT_ROUTE.value:
                if default_route:
                    raise q_exc.DuplicateRouteConflict(
                        subnet_id=new_subnet["id"])

                default_route = route
                gateway_ip = default_route["nexthop"]
                alloc_pools.validate_gateway_excluded(gateway_ip)

            new_subnet["routes"].append(db_api.route_create(
                context, cidr=route["destination"], gateway=route["nexthop"]))

        quota.QUOTAS.limit_check(context, context.tenant_id,
                                 dns_nameservers_per_subnet=len(dns_ips))

        for dns_ip in dns_ips:
            new_subnet["dns_nameservers"].append(db_api.dns_create(
                context, ip=netaddr.IPAddress(dns_ip)))

        # if the gateway_ip is IN the cidr for the subnet and NOT excluded by
        # policies, we should raise a 409 conflict
        if gateway_ip and default_route is None:
            alloc_pools.validate_gateway_excluded(gateway_ip)
            new_subnet["routes"].append(db_api.route_create(
                context, cidr=str(routes.DEFAULT_ROUTE), gateway=gateway_ip))

    subnet_dict = v._make_subnet_dict(new_subnet)
    subnet_dict["gateway_ip"] = gateway_ip

    return subnet_dict


def _pool_is_growing(original_pool, new_pool):
    # create IPSet for original pool
    ori_set = netaddr.IPSet()
    for rng in original_pool._alloc_pools:
        ori_set.add(netaddr.IPRange(rng['start'], rng['end']))

    # create IPSet for net pool
    new_set = netaddr.IPSet()
    for rng in new_pool._alloc_pools:
        new_set.add(netaddr.IPRange(rng['start'], rng['end']))

    # we are growing the original set is not a superset of the new set
    return not ori_set.issuperset(new_set)


def update_subnet(context, id, subnet):
    """Update values of a subnet.

    : param context: neutron api request context
    : param id: UUID representing the subnet to update.
    : param subnet: dictionary with keys indicating fields to update.
        valid keys are those that have a value of True for 'allow_put'
        as listed in the RESOURCE_ATTRIBUTE_MAP object in
        neutron/api/v2/attributes.py.
    """
    LOG.info("update_subnet %s for tenant %s" %
             (id, context.tenant_id))

    with context.session.begin():
        subnet_db = db_api.subnet_find(context, None, None, None, False, id=id,
                                       scope=db_api.ONE)
        if not subnet_db:
            raise n_exc.SubnetNotFound(subnet_id=id)

        s = subnet["subnet"]
        always_pop = ["_cidr", "cidr", "first_ip", "last_ip", "ip_version",
                      "segment_id", "network_id"]
        admin_only = ["do_not_use", "created_at", "tenant_id",
                      "next_auto_assign_ip", "enable_dhcp"]
        utils.filter_body(context, s, admin_only, always_pop)

        dns_ips = utils.pop_param(s, "dns_nameservers", [])
        host_routes = utils.pop_param(s, "host_routes", [])
        gateway_ip = utils.pop_param(s, "gateway_ip", None)
        allocation_pools = utils.pop_param(s, "allocation_pools", None)
        if not CONF.QUARK.allow_allocation_pool_update:
            if allocation_pools:
                raise n_exc.BadRequest(
                    resource="subnets",
                    msg="Allocation pools cannot be updated.")

            if subnet_db["ip_policy"] is not None:
                ip_policy_cidrs = subnet_db["ip_policy"].get_cidrs_ip_set()
            else:
                ip_policy_cidrs = netaddr.IPSet([])

            alloc_pools = allocation_pool.AllocationPools(
                subnet_db["cidr"],
                policies=ip_policy_cidrs)
        else:
            alloc_pools = allocation_pool.AllocationPools(subnet_db["cidr"],
                                                          allocation_pools)
            original_pools = subnet_db.allocation_pools
            ori_pools = allocation_pool.AllocationPools(subnet_db["cidr"],
                                                        original_pools)
            # Check if the pools are growing or shrinking
            is_growing = _pool_is_growing(ori_pools, alloc_pools)
            if not CONF.QUARK.allow_allocation_pool_growth and is_growing:
                raise n_exc.BadRequest(
                    resource="subnets",
                    msg="Allocation pools may not be updated to be larger "
                        "do to configuration settings")

        quota.QUOTAS.limit_check(
            context,
            context.tenant_id,
            alloc_pools_per_subnet=len(alloc_pools))
        if gateway_ip:
            alloc_pools.validate_gateway_excluded(gateway_ip)
            default_route = None
            for route in host_routes:
                netaddr_route = netaddr.IPNetwork(route["destination"])
                if netaddr_route.value == routes.DEFAULT_ROUTE.value:
                    default_route = route
                    break

            if default_route is None:
                route_model = db_api.route_find(
                    context, cidr=str(routes.DEFAULT_ROUTE), subnet_id=id,
                    scope=db_api.ONE)
                if route_model:
                    db_api.route_update(context, route_model,
                                        gateway=gateway_ip)
                else:
                    db_api.route_create(context,
                                        cidr=str(routes.DEFAULT_ROUTE),
                                        gateway=gateway_ip, subnet_id=id)

        if dns_ips:
            subnet_db["dns_nameservers"] = []
            quota.QUOTAS.limit_check(context, context.tenant_id,
                                     dns_nameservers_per_subnet=len(dns_ips))

        for dns_ip in dns_ips:
            subnet_db["dns_nameservers"].append(db_api.dns_create(
                context,
                ip=netaddr.IPAddress(dns_ip)))

        if host_routes:
            subnet_db["routes"] = []
            quota.QUOTAS.limit_check(context, context.tenant_id,
                                     routes_per_subnet=len(host_routes))

        for route in host_routes:
            subnet_db["routes"].append(db_api.route_create(
                context, cidr=route["destination"], gateway=route["nexthop"]))
        if CONF.QUARK.allow_allocation_pool_update:
            if isinstance(allocation_pools, list):
                cidrs = alloc_pools.get_policy_cidrs()
                ip_policies.ensure_default_policy(cidrs, [subnet_db])
                subnet_db["ip_policy"] = db_api.ip_policy_update(
                    context, subnet_db["ip_policy"], exclude=cidrs)
                # invalidate the cache
                db_api.subnet_update_set_alloc_pool_cache(context, subnet_db)
        subnet = db_api.subnet_update(context, subnet_db, **s)
    return v._make_subnet_dict(subnet)


def get_subnet(context, id, fields=None):
    """Retrieve a subnet.

    : param context: neutron api request context
    : param id: UUID representing the subnet to fetch.
    : param fields: a list of strings that are valid keys in a
        subnet dictionary as listed in the RESOURCE_ATTRIBUTE_MAP
        object in neutron/api/v2/attributes.py. Only these fields
        will be returned.
    """
    LOG.info("get_subnet %s for tenant %s with fields %s" %
             (id, context.tenant_id, fields))
    subnet = db_api.subnet_find(context, None, None, None, False, id=id,
                                join_dns=True, join_routes=True,
                                scope=db_api.ONE)
    if not subnet:
        raise n_exc.SubnetNotFound(subnet_id=id)

    cache = subnet.get("_allocation_pool_cache")
    if not cache:
        new_cache = subnet.allocation_pools
        db_api.subnet_update_set_alloc_pool_cache(context, subnet, new_cache)
    return v._make_subnet_dict(subnet)


def get_subnets(context, limit=None, page_reverse=False, sorts=None,
                marker=None, filters=None, fields=None):
    """Retrieve a list of subnets.

    The contents of the list depends on the identity of the user
    making the request (as indicated by the context) as well as any
    filters.
    : param context: neutron api request context
    : param filters: a dictionary with keys that are valid keys for
        a subnet as listed in the RESOURCE_ATTRIBUTE_MAP object
        in neutron/api/v2/attributes.py.  Values in this dictiontary
        are an iterable containing values that will be used for an exact
        match comparison for that value.  Each result returned by this
        function will have matched one of the values for each key in
        filters.
    : param fields: a list of strings that are valid keys in a
        subnet dictionary as listed in the RESOURCE_ATTRIBUTE_MAP
        object in neutron/api/v2/attributes.py. Only these fields
        will be returned.
    """
    LOG.info("get_subnets for tenant %s with filters %s fields %s" %
             (context.tenant_id, filters, fields))
    filters = filters or {}
    subnets = db_api.subnet_find(context, limit=limit,
                                 page_reverse=page_reverse, sorts=sorts,
                                 marker_obj=marker, join_dns=True,
                                 join_routes=True, join_pool=True, **filters)
    for subnet in subnets:
        cache = subnet.get("_allocation_pool_cache")
        if not cache:
            db_api.subnet_update_set_alloc_pool_cache(
                context, subnet, subnet.allocation_pools)
    return v._make_subnets_list(subnets, fields=fields)


def get_subnets_count(context, filters=None):
    """Return the number of subnets.

    The result depends on the identity of the user making the request
    (as indicated by the context) as well as any filters.
    : param context: neutron api request context
    : param filters: a dictionary with keys that are valid keys for
        a network as listed in the RESOURCE_ATTRIBUTE_MAP object
        in neutron/api/v2/attributes.py.  Values in this dictiontary
        are an iterable containing values that will be used for an exact
        match comparison for that value.  Each result returned by this
        function will have matched one of the values for each key in
        filters.

    NOTE: this method is optional, as it was not part of the originally
          defined plugin API.
    """
    LOG.info("get_subnets_count for tenant %s with filters %s" %
             (context.tenant_id, filters))
    return db_api.subnet_count_all(context, **filters)


def _delete_subnet(context, subnet):
    if subnet.allocated_ips:
        raise n_exc.SubnetInUse(subnet_id=subnet["id"])
    db_api.subnet_delete(context, subnet)


def delete_subnet(context, id):
    """Delete a subnet.

    : param context: neutron api request context
    : param id: UUID representing the subnet to delete.
    """
    LOG.info("delete_subnet %s for tenant %s" % (id, context.tenant_id))
    with context.session.begin():
        subnet = db_api.subnet_find(context, id=id, scope=db_api.ONE)
        if not subnet:
            raise n_exc.SubnetNotFound(subnet_id=id)

        if not context.is_admin:
            if STRATEGY.is_provider_network(subnet.network_id):
                if subnet.tenant_id == context.tenant_id:
                    # A tenant can't delete subnets on provider network
                    raise n_exc.NotAuthorized(subnet_id=id)
                else:
                    # Raise a NotFound here because the foreign tenant
                    # does not have to know about other tenant's subnet
                    # existence.
                    raise n_exc.SubnetNotFound(subnet_id=id)

        _delete_subnet(context, subnet)


def diagnose_subnet(context, id, fields):
    if not context.is_admin:
        raise n_exc.NotAuthorized()

    if id == "*":
        return {'subnets': get_subnets(context, filters={})}
    return {'subnets': get_subnet(context, id)}



# Copyright (c) 2015 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import contextlib

from oslo_config import cfg

from quark import environment as env
from quark.tests import test_base


class TestEnvironment(test_base.TestBase):
    def setUp(self):
        super(TestEnvironment, self).setUp()
        self.MAGIC_VALUE = "12345"

    @contextlib.contextmanager
    def _fixture(self, has_capability):
        old_override = cfg.CONF.QUARK.environment_capabilities
        if has_capability:
            override = env.Capabilities.SECURITY_GROUPS
        else:
            override = ""

        cfg.CONF.set_override("environment_capabilities",
                              override,
                              "QUARK")
        yield

        cfg.CONF.set_override("environment_capabilities",
                              old_override,
                              "QUARK")

    @env.has_capability(env.Capabilities.SECURITY_GROUPS)
    def foo(self):
        return self.MAGIC_VALUE

    def test_has_capability_True(self):
        with self._fixture(True):
            self.assertEqual(self.foo(), self.MAGIC_VALUE)

    def test_has_capability_False(self):
        with self._fixture(False):
            self.assertIsNone(self.foo())



# Copyright 2013 Rackspace Hosting Inc.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
#  under the License.

import contextlib
import copy
from datetime import datetime
import json
import uuid

import mock
from neutron.api.v2 import attributes as neutron_attrs
from neutron.common import exceptions as n_exc_ext
from neutron_lib import exceptions as n_exc
from oslo_config import cfg

from quark.db import api as db_api
from quark.db import models
from quark import exceptions as q_exc
from quark import network_strategy
from quark import plugin_views
from quark.tests import test_quark_plugin


class TestQuarkGetSubnetCount(test_quark_plugin.TestQuarkPlugin):
    def test_get_subnet_count(self):
        """This isn't really testable."""
        with mock.patch("quark.db.api.subnet_count_all"):
            self.plugin.get_subnets_count(self.context, {})


class TestQuarkGetSubnets(test_quark_plugin.TestQuarkPlugin):
    @contextlib.contextmanager
    def _stubs(self, subnets=None, routes=None):
        if routes is None:
            routes = []
        route_models = []
        for route in routes:
            r = models.Route()
            r.update(route)
            route_models.append(r)

        if isinstance(subnets, list):
            subnet_models = []
            for subnet in subnets:
                s_dict = subnet.copy()
                s_dict["routes"] = route_models
                s = models.Subnet(network=models.Network())
                s.update(s_dict)
                subnet_models.append(s)
        elif subnets:
            mod = models.Subnet(network=models.Network())
            mod.update(subnets)
            mod["routes"] = route_models
            subnet_models = mod
        else:
            subnet_models = None

        with mock.patch("quark.db.api.subnet_find") as subnet_find:
            subnet_find.return_value = subnet_models
            yield

    def test_subnets_list(self):
        subnet_id = str(uuid.uuid4())
        route = dict(id=1, cidr="0.0.0.0/0", gateway="192.168.0.1")

        subnet = dict(id=subnet_id, network_id=1, name=subnet_id,
                      tenant_id=self.context.tenant_id, ip_version=4,
                      cidr="192.168.0.0/24", gateway_ip="192.168.0.1",
                      dns_nameservers=[],
                      enable_dhcp=None)

        with self._stubs(subnets=[subnet], routes=[route]):
            res = self.plugin.get_subnets(self.context, None, None, None, {},
                                          {})
            # Compare routes separately
            routes = res[0].pop("host_routes")
            for key in subnet.keys():
                self.assertEqual(res[0][key], subnet[key])
            self.assertEqual(len(routes), 0)

    def test_subnets_list_two_default_routes_shows_last_one(self):
        subnet_id = str(uuid.uuid4())
        route = dict(id=1, cidr="0.0.0.0/0", gateway="192.168.0.1")
        route2 = dict(id=1, cidr="0.0.0.0/0", gateway="192.168.0.2")

        subnet = dict(id=subnet_id, network_id=1, name=subnet_id,
                      tenant_id=self.context.tenant_id, ip_version=4,
                      cidr="192.168.0.0/24", gateway_ip="192.168.0.1",
                      dns_nameservers=[],
                      enable_dhcp=None)

        with self._stubs(subnets=[subnet], routes=[route, route2]):
            res = self.plugin.get_subnets(self.context, None, None, None, {},
                                          {})

            # Don't want to test that LOG.info is called but we can
            # know the case is covered by checking the gateway is the one
            # we expect it to be
            self.assertEqual(res[0]["gateway_ip"], "192.168.0.2")
            self.assertEqual(len(res[0]["host_routes"]), 0)

    def test_subnet_show_fail(self):
        with self._stubs():
            with self.assertRaises(n_exc.SubnetNotFound):
                self.plugin.get_subnet(self.context, 1)

    def test_subnet_show(self):
        subnet_id = str(uuid.uuid4())
        route = dict(id=1, cidr="0.0.0.0/0", gateway="192.168.0.1",
                     subnet_id=subnet_id)

        subnet = dict(id=subnet_id, network_id=1, name=subnet_id,
                      tenant_id=self.context.tenant_id, ip_version=4,
                      cidr="192.168.0.0/24", gateway_ip="192.168.0.1",
                      dns_nameservers=[],
                      enable_dhcp=None)

        with self._stubs(subnets=subnet, routes=[route]):
            res = self.plugin.get_subnet(self.context, subnet_id)

            # Compare routes separately
            routes = res.pop("host_routes")
            for key in subnet.keys():
                self.assertEqual(res[key], subnet[key])
            self.assertEqual(len(routes), 0)


class TestQuarkGetSubnetsHideAllocPools(test_quark_plugin.TestQuarkPlugin):
    @contextlib.contextmanager
    def _stubs(self, subnets=None):
        if isinstance(subnets, list):
            subnet_models = []
            for subnet in subnets:
                s_dict = subnet.copy()
                s = models.Subnet(network=models.Network())
                s.update(s_dict)
                subnet_models.append(s)

        cfg.CONF.set_override('show_allocation_pools', False, "QUARK")
        with mock.patch("quark.db.api.subnet_find") as subnet_find:
            subnet_find.return_value = subnet_models
            yield
        cfg.CONF.set_override('show_allocation_pools', True, "QUARK")

    def test_subnets_list(self):
        subnet_id = str(uuid.uuid4())

        subnet = dict(id=subnet_id, network_id=1, name=subnet_id,
                      tenant_id=self.context.tenant_id, ip_version=4,
                      cidr="192.168.0.0/24", gateway_ip="192.168.0.1",
                      dns_nameservers=[],
                      enable_dhcp=None)

        with self._stubs(subnets=[subnet]):
            res = self.plugin.get_subnets(self.context, None, None, None, {},
                                          {})
            self.assertEqual(res[0]["allocation_pools"], [])


class TestQuarkCreateSubnetOverlapping(test_quark_plugin.TestQuarkPlugin):
    @contextlib.contextmanager
    def _stubs(self, subnets=None):
        if subnets is None:
            subnets = []
        subnet_models = []
        for subnet in subnets:
            s = models.Subnet()
            s.update(subnet)
            subnet_models.append(s)
        network = models.Network()
        network.update(dict(id=1, subnets=subnet_models))
        with contextlib.nested(
            mock.patch("quark.db.api.network_find"),
            mock.patch("quark.db.api.subnet_find"),
            mock.patch("quark.db.api.subnet_create"),
            mock.patch("neutron.common.rpc.get_notifier")
        ) as (net_find, subnet_find, subnet_create, get_notifier):
            net_find.return_value = network
            subnet_find.return_value = subnet_models
            subnet_create.return_value = models.Subnet(
                network=models.Network(),
                cidr="192.168.1.1/24")
            yield subnet_create

    def test_create_subnet_overlapping_true(self):
        cfg.CONF.set_override('allow_overlapping_ips', True)
        with self._stubs() as subnet_create:
            s = dict(subnet=dict(
                gateway_ip=neutron_attrs.ATTR_NOT_SPECIFIED,
                dns_nameservers=neutron_attrs.ATTR_NOT_SPECIFIED,
                cidr="192.168.1.1/8",
                network_id=1))
            self.plugin.create_subnet(self.context, s)
            self.assertEqual(subnet_create.call_count, 1)

    def test_create_subnet_overlapping_false(self):
        cfg.CONF.set_override('allow_overlapping_ips', False)
        with self._stubs() as subnet_create:
            s = dict(subnet=dict(
                gateway_ip=neutron_attrs.ATTR_NOT_SPECIFIED,
                dns_nameservers=neutron_attrs.ATTR_NOT_SPECIFIED,
                cidr="192.168.1.1/8",
                network_id=1))
            self.plugin.create_subnet(self.context, s)
            self.assertEqual(subnet_create.call_count, 1)

    def test_create_subnet_overlapping_conflict(self):
        cfg.CONF.set_override('allow_overlapping_ips', False)
        with self._stubs(subnets=[dict(cidr="192.168.10.1/24")]):
            with self.assertRaises(n_exc.InvalidInput):
                s = dict(subnet=dict(cidr="192.168.1.1/8",
                                     network_id=1))
                self.plugin.create_subnet(self.context, s)


class TestQuarkCreateSubnetAllocationPools(test_quark_plugin.TestQuarkPlugin):
    @contextlib.contextmanager
    def _stubs(self, subnet):
        s = models.Subnet(network=models.Network(id=1, subnets=[]))
        allocation_pools = subnet.pop("allocation_pools", None)
        s.update(subnet)
        if allocation_pools is not None:
            subnet["allocation_pools"] = allocation_pools

        def _allocation_pools_mock():
            if allocation_pools is not None:
                return mock.patch.object(models.Subnet, "allocation_pools")
            return mock.MagicMock()

        with contextlib.nested(
            mock.patch("quark.db.api.network_find"),
            mock.patch("quark.db.api.subnet_find"),
            mock.patch("quark.db.api.subnet_create"),
            mock.patch("neutron.common.rpc.get_notifier"),
            _allocation_pools_mock(),
        ) as (net_find, subnet_find, subnet_create, get_notifier,
              alloc_pools_method):
            net_find.return_value = s["network"]
            subnet_find.return_value = []
            subnet_create.return_value = s
            alloc_pools_method.__get__ = mock.Mock(
                return_value=allocation_pools)
            yield subnet_create

    def setUp(self):
        super(TestQuarkCreateSubnetAllocationPools, self).setUp()

    def tearDown(self):
        super(TestQuarkCreateSubnetAllocationPools, self).tearDown()

    def test_create_subnet_allocation_pools_zero(self):
        s = dict(subnet=dict(
            cidr="192.168.1.1/24",
            network_id=1))
        with self._stubs(s["subnet"]) as (subnet_create):
            resp = self.plugin.create_subnet(self.context, s)
            self.assertEqual(subnet_create.call_count, 1)
            self.assertEqual(resp["allocation_pools"],
                             [dict(start="192.168.1.1", end="192.168.1.254")])

    def test_create_subnet_allocation_pools_zero_v6(self):
        s = dict(subnet=dict(
            cidr="2607:f0d0:1002:51::0/64",
            network_id=1))
        with self._stubs(s["subnet"]) as (subnet_create):
            resp = self.plugin.create_subnet(self.context, s)
            self.assertEqual(subnet_create.call_count, 1)
            self.assertEqual(
                resp["allocation_pools"],
                [dict(start="2607:f0d0:1002:51::1",
                      end="2607:f0d0:1002:51:ffff:ffff:ffff:fffe")])

    def test_create_subnet_allocation_pools_one(self):
        pools = [dict(start="192.168.1.10", end="192.168.1.20")]
        s = dict(subnet=dict(
            allocation_pools=pools,
            cidr="192.168.1.1/24",
            network_id=1))
        with self._stubs(s["subnet"]) as (subnet_create):
            resp = self.plugin.create_subnet(self.context, s)
            self.assertEqual(subnet_create.call_count, 1)
            self.assertEqual(resp["allocation_pools"], pools)

    def test_create_subnet_allocation_pools_gateway_conflict(self):
        pools = [dict(start="192.168.1.1", end="192.168.1.20")]
        s = dict(subnet=dict(allocation_pools=pools,
                             cidr="192.168.1.1/24",
                             gateway_ip="192.168.1.1",
                             network_id=1))
        with self._stubs(s["subnet"]):
            with self.assertRaises(
                    n_exc_ext.GatewayConflictWithAllocationPools):
                self.plugin.create_subnet(self.context, s)

    def test_create_subnet_allocation_pools_invalid_outside(self):
        pools = [dict(start="192.168.0.10", end="192.168.0.20")]
        s = dict(subnet=dict(
            allocation_pools=pools,
            cidr="192.168.1.1/24",
            network_id=1))
        with self._stubs(s["subnet"]):
            with self.assertRaises(n_exc_ext.OutOfBoundsAllocationPool):
                self.plugin.create_subnet(self.context, s)

    def test_create_subnet_allocation_pools_invalid_overlaps(self):
        pools = [dict(start="192.168.0.255", end="192.168.1.20")]
        s = dict(subnet=dict(
            allocation_pools=pools,
            cidr="192.168.1.1/24",
            network_id=1))
        with self._stubs(s["subnet"]):
            with self.assertRaises(n_exc_ext.OutOfBoundsAllocationPool):
                self.plugin.create_subnet(self.context, s)

    def test_create_subnet_allocation_pools_two(self):
        pools = [dict(start="192.168.1.10", end="192.168.1.20"),
                 dict(start="192.168.1.40", end="192.168.1.50")]
        s = dict(subnet=dict(
            allocation_pools=pools,
            cidr="192.168.1.1/24",
            network_id=1))
        with self._stubs(s["subnet"]) as (subnet_create):
            resp = self.plugin.create_subnet(self.context, s)
            self.assertEqual(subnet_create.call_count, 1)
            self.assertEqual(resp["allocation_pools"], pools)

    def test_create_subnet_allocation_pools_three(self):
        pools = [dict(start="192.168.1.5", end="192.168.1.254")]
        s = dict(subnet=dict(
            allocation_pools=pools,
            ip_version=4,
            cidr="192.168.1.1/24",
            network_id=1))
        with self._stubs(s["subnet"]) as (subnet_create):
            resp = self.plugin.create_subnet(self.context, s)
            self.assertEqual(subnet_create.call_count, 1)
            self.assertEqual(resp["allocation_pools"], pools)

    def test_create_subnet_allocation_pools_four(self):
        pools = [dict(start="2607:f0d0:1002:51::a",
                 end="2607:f0d0:1002:51::ffff:fffe")]
        s = dict(subnet=dict(
            allocation_pools=pools,
            ip_version=6,
            cidr="2607:f0d0:1002:51::0/64",
            network_id=1))
        with self._stubs(s["subnet"]) as (subnet_create):
            resp = self.plugin.create_subnet(self.context, s)
            self.assertEqual(subnet_create.call_count, 1)
            self.assertEqual(resp["allocation_pools"], pools)

    def test_create_subnet_allocation_pools_empty_list(self):
        # Empty allocation_pools list yields subnet completely blocked out.
        pools = []
        s = dict(subnet=dict(
            allocation_pools=pools,
            cidr="192.168.1.1/24",
            network_id=1))
        with self._stubs(s["subnet"]) as (subnet_create):
            resp = self.plugin.create_subnet(self.context, s)
            self.assertEqual(subnet_create.call_count, 1)
            expected_pools = []
            self.assertEqual(resp["allocation_pools"], expected_pools)


# TODO(amir): Refactor the tests to test individual subnet attributes.
# * copy.deepcopy was necessary to maintain tests on keys, which is a bit ugly.
# * workaround is also in place for lame ATTR_NOT_SPECIFIED object()
class TestQuarkCreateSubnet(test_quark_plugin.TestQuarkPlugin):
    @contextlib.contextmanager
    def _stubs(self, subnet=None, network=True, routes=(), dns=()):

        subnet_mod = models.Subnet(
            network=models.Network(id=1) if network else None)
        dns_ips = subnet.pop("dns_nameservers", [])
        host_routes = subnet.pop("host_routes", [])
        allocation_pools = subnet.pop("allocation_pools", None)
        subnet_mod.update(subnet)

        subnet["dns_nameservers"] = dns_ips
        subnet["host_routes"] = host_routes
        if allocation_pools is not None:
            subnet["allocation_pools"] = allocation_pools
        dns = [{"ip": x} for x in dns]
        route_models = [models.Route(**r) for r in routes]
        dns_models = [models.DNSNameserver(**d) for d in dns]

        def _allocation_pools_mock():
            if allocation_pools is not None:
                return mock.patch.object(models.Subnet, "allocation_pools")
            return mock.MagicMock()
        with contextlib.nested(
            mock.patch("quark.db.api.subnet_create"),
            mock.patch("quark.db.api.network_find"),
            mock.patch("quark.db.api.dns_create"),
            mock.patch("quark.db.api.route_create"),
            mock.patch("quark.db.api.subnet_find"),
            mock.patch("neutron.common.rpc.get_notifier"),
            _allocation_pools_mock()
        ) as (subnet_create, net_find, dns_create, route_create, subnet_find,
              get_notifier, alloc_pools_method):
            subnet_create.return_value = subnet_mod
            net_find.return_value = network
            route_create.side_effect = route_models
            dns_create.side_effect = dns_models
            alloc_pools_method.__get__ = mock.Mock(
                return_value=allocation_pools)
            yield subnet_create, dns_create, route_create

    def test_create_subnet(self):
        routes = [dict(cidr="0.0.0.0/0", gateway="0.0.0.0")]
        subnet = dict(
            subnet=dict(network_id=1,
                        tenant_id=self.context.tenant_id, ip_version=4,
                        cidr="172.16.0.0/24", gateway_ip="0.0.0.0",
                        dns_nameservers=neutron_attrs.ATTR_NOT_SPECIFIED,
                        host_routes=neutron_attrs.ATTR_NOT_SPECIFIED,
                        enable_dhcp=None))
        with self._stubs(
            subnet=subnet["subnet"],
            routes=routes
        ) as (subnet_create, dns_create, route_create):
            dns_nameservers = subnet["subnet"].pop("dns_nameservers")
            host_routes = subnet["subnet"].pop("host_routes")
            subnet_request = copy.deepcopy(subnet)
            subnet_request["subnet"]["dns_nameservers"] = dns_nameservers
            subnet_request["subnet"]["host_routes"] = host_routes
            res = self.plugin.create_subnet(self.context,
                                            subnet_request)
            self.assertEqual(subnet_create.call_count, 1)
            self.assertEqual(dns_create.call_count, 0)
            self.assertEqual(route_create.call_count, 1)
            for key in subnet["subnet"].keys():
                if key == "host_routes":
                    self.assertEqual(res[key][0]["destination"], "0.0.0.0/0")
                    self.assertEqual(res[key][0]["nexthop"], "0.0.0.0")
                else:
                    self.assertEqual(res[key], subnet["subnet"][key])
            expected_pools = [{'start': '172.16.0.1',
                              'end': '172.16.0.254'}]
            self.assertEqual(res["allocation_pools"], expected_pools)

    def test_create_subnet_v6_too_small(self):
        routes = [dict(cidr="0.0.0.0/0", gateway="0.0.0.0")]
        subnet = dict(
            subnet=dict(network_id=1,
                        tenant_id=self.context.tenant_id, ip_version=4,
                        cidr="1234::/80", gateway_ip="0.0.0.0",
                        dns_nameservers=neutron_attrs.ATTR_NOT_SPECIFIED,
                        host_routes=neutron_attrs.ATTR_NOT_SPECIFIED,
                        enable_dhcp=None))
        with self._stubs(
            subnet=subnet["subnet"],
            routes=routes
        ) as (subnet_create, dns_create, route_create):
            dns_nameservers = subnet["subnet"].pop("dns_nameservers")
            host_routes = subnet["subnet"].pop("host_routes")
            subnet_request = copy.deepcopy(subnet)
            subnet_request["subnet"]["dns_nameservers"] = dns_nameservers
            subnet_request["subnet"]["host_routes"] = host_routes
            with self.assertRaises(n_exc.InvalidInput):
                self.plugin.create_subnet(self.context, subnet_request)

    def test_create_subnet_v4_too_small(self):
        routes = [dict(cidr="0.0.0.0/0", gateway="0.0.0.0")]
        subnet = dict(
            subnet=dict(network_id=1,
                        tenant_id=self.context.tenant_id, ip_version=4,
                        cidr="192.168.0.0/31", gateway_ip="0.0.0.0",
                        dns_nameservers=neutron_attrs.ATTR_NOT_SPECIFIED,
                        host_routes=neutron_attrs.ATTR_NOT_SPECIFIED,
                        enable_dhcp=None))
        with self._stubs(
            subnet=subnet["subnet"],
            routes=routes
        ) as (subnet_create, dns_create, route_create):
            dns_nameservers = subnet["subnet"].pop("dns_nameservers")
            host_routes = subnet["subnet"].pop("host_routes")
            subnet_request = copy.deepcopy(subnet)
            subnet_request["subnet"]["dns_nameservers"] = dns_nameservers
            subnet_request["subnet"]["host_routes"] = host_routes
            with self.assertRaises(n_exc.InvalidInput):
                self.plugin.create_subnet(self.context, subnet_request)

    def test_create_subnet_not_admin_segment_id_ignored(self):
        routes = [dict(cidr="0.0.0.0/0", gateway="0.0.0.0")]
        subnet = dict(
            subnet=dict(network_id=1,
                        tenant_id=self.context.tenant_id, ip_version=4,
                        cidr="172.16.0.0/24", gateway_ip="0.0.0.0",
                        dns_nameservers=neutron_attrs.ATTR_NOT_SPECIFIED,
                        host_routes=neutron_attrs.ATTR_NOT_SPECIFIED,
                        enable_dhcp=None))
        with self._stubs(
            subnet=subnet["subnet"],
            routes=routes
        ) as (subnet_create, dns_create, route_create):
            dns_nameservers = subnet["subnet"].pop("dns_nameservers")
            host_routes = subnet["subnet"].pop("host_routes")
            subnet_request = copy.deepcopy(subnet)
            subnet_request["subnet"]["dns_nameservers"] = dns_nameservers
            subnet_request["subnet"]["host_routes"] = host_routes
            subnet_request["subnet"]["segment_id"] = "cell01"
            res = self.plugin.create_subnet(self.context,
                                            subnet_request)
            self.assertEqual(subnet_create.call_count, 1)
            self.assertTrue("segment_id" not in subnet_create.called_with)

            self.assertEqual(dns_create.call_count, 0)
            self.assertEqual(route_create.call_count, 1)
            for key in subnet["subnet"].keys():
                if key == "host_routes":
                    self.assertEqual(res[key][0]["destination"], "0.0.0.0/0")
                    self.assertEqual(res[key][0]["nexthop"], "0.0.0.0")
                else:
                    self.assertEqual(res[key], subnet["subnet"][key])

    def test_create_subnet_no_network_fails(self):
        subnet = dict(subnet=dict(network_id=1))
        with self._stubs(subnet=dict(), network=False):
            with self.assertRaises(n_exc.NetworkNotFound):
                self.plugin.create_subnet(self.context, subnet)

    def test_create_subnet_no_gateway_ip_defaults(self):
        subnet = dict(
            subnet=dict(network_id=1,
                        tenant_id=self.context.tenant_id, ip_version=4,
                        cidr="172.16.0.0/24",
                        gateway_ip=neutron_attrs.ATTR_NOT_SPECIFIED,
                        dns_nameservers=neutron_attrs.ATTR_NOT_SPECIFIED,
                        enable_dhcp=None))
        with self._stubs(
            subnet=subnet["subnet"],
            routes=[]
        ) as (subnet_create, dns_create, route_create):
            dns_nameservers = subnet["subnet"].pop("dns_nameservers")
            gateway_ip = subnet["subnet"].pop("gateway_ip")
            subnet_request = copy.deepcopy(subnet)
            subnet_request["subnet"]["dns_nameservers"] = dns_nameservers
            subnet_request["subnet"]["gateway_ip"] = gateway_ip
            res = self.plugin.create_subnet(self.context, subnet_request)
            self.assertEqual(subnet_create.call_count, 1)
            self.assertEqual(dns_create.call_count, 0)
            self.assertEqual(route_create.call_count, 0)
            for key in subnet["subnet"].keys():
                if key == "gateway_ip":
                    self.assertEqual(res[key], "172.16.0.1")
                elif key == "host_routes":
                    self.assertEqual(len(res[key]), 0)
                else:
                    self.assertEqual(res[key], subnet["subnet"][key])

    def test_create_subnet_dns_nameservers(self):
        routes = [dict(cidr="0.0.0.0/0", gateway="0.0.0.0")]
        dns_ns = ["4.2.2.1", "4.2.2.2"]
        subnet = dict(
            subnet=dict(network_id=1,
                        tenant_id=self.context.tenant_id, ip_version=4,
                        cidr="172.16.0.0/24", gateway_ip="0.0.0.0",
                        dns_nameservers=dns_ns, enable_dhcp=None))
        with self._stubs(
            subnet=subnet["subnet"],
            routes=routes,
            dns=dns_ns
        ) as (subnet_create, dns_create, route_create):
            res = self.plugin.create_subnet(self.context,
                                            copy.deepcopy(subnet))
            self.assertEqual(subnet_create.call_count, 1)
            self.assertEqual(dns_create.call_count, 2)
            self.assertEqual(route_create.call_count, 1)
            for key in subnet["subnet"].keys():
                if key == "host_routes":
                    self.assertEqual(len(res[key]), 0)
                else:
                    self.assertEqual(res[key], subnet["subnet"][key])

    def test_create_subnet_routes(self):
        routes = [dict(cidr="1.1.1.1/8", gateway="172.16.0.4"),
                  dict(cidr="0.0.0.0/0", gateway="0.0.0.0")]
        subnet = dict(
            subnet=dict(network_id=1,
                        tenant_id=self.context.tenant_id, ip_version=4,
                        cidr="172.16.0.0/24", gateway_ip="0.0.0.0",
                        dns_nameservers=neutron_attrs.ATTR_NOT_SPECIFIED,
                        host_routes=[{"destination": "1.1.1.1/8",
                                      "nexthop": "172.16.0.4"}],
                        allocation_pools=[{"start": "172.16.0.5",
                                           "end": "172.16.0.254"}],
                        enable_dhcp=None))
        with self._stubs(
            subnet=subnet["subnet"],
            routes=routes
        ) as (subnet_create, dns_create, route_create):
            dns_nameservers = subnet["subnet"].pop("dns_nameservers")
            subnet_request = copy.deepcopy(subnet)
            subnet_request["subnet"]["dns_nameservers"] = dns_nameservers
            res = self.plugin.create_subnet(self.context, subnet_request)
            self.assertEqual(subnet_create.call_count, 1)
            self.assertEqual(dns_create.call_count, 0)
            self.assertEqual(route_create.call_count, 2)
            for key in subnet["subnet"].keys():
                if key == "host_routes":
                    res_tuples = [(r["destination"], r["nexthop"])
                                  for r in res[key]]
                    self.assertIn(("1.1.1.1/8", "172.16.0.4"), res_tuples)
                    self.assertEqual(1, len(res_tuples))
                else:
                    self.assertEqual(res[key], subnet["subnet"][key])

    def test_create_subnet_default_route(self):
        routes = [dict(cidr="0.0.0.0/0", gateway="172.16.0.4")]
        subnet = dict(
            subnet=dict(network_id=1,
                        tenant_id=self.context.tenant_id, ip_version=4,
                        cidr="172.16.0.0/24",
                        gateway_ip=neutron_attrs.ATTR_NOT_SPECIFIED,
                        dns_nameservers=neutron_attrs.ATTR_NOT_SPECIFIED,
                        host_routes=[{"destination": "0.0.0.0/0",
                                      "nexthop": "172.16.0.4"}],
                        allocation_pools=[{"start": "172.16.0.5",
                                           "end": "172.16.0.254"}],
                        enable_dhcp=None))
        with self._stubs(
            subnet=subnet["subnet"],
            routes=routes
        ) as (subnet_create, dns_create, route_create):
            dns_nameservers = subnet["subnet"].pop("dns_nameservers")
            gateway_ip = subnet["subnet"].pop("gateway_ip")
            subnet_request = copy.deepcopy(subnet)
            subnet_request["subnet"]["dns_nameservers"] = dns_nameservers
            subnet_request["subnet"]["gateway_ip"] = gateway_ip
            res = self.plugin.create_subnet(self.context, subnet_request)
            self.assertEqual(subnet_create.call_count, 1)
            self.assertEqual(dns_create.call_count, 0)
            self.assertEqual(route_create.call_count, 1)
            for key in subnet["subnet"].keys():
                if key == "gateway_ip":
                    self.assertEqual(res[key], "172.16.0.4")
                elif key == "host_routes":
                    self.assertEqual(len(res[key]), 0)
                else:
                    self.assertEqual(res[key], subnet["subnet"][key])

    def test_create_subnet_two_default_routes_fails(self):
        routes = [dict(cidr="0.0.0.0/0", gateway="172.16.0.4"),
                  dict(cidr="0.0.0.0/0", gateway="172.16.0.4")]
        subnet = dict(
            subnet=dict(network_id=1,
                        tenant_id=self.context.tenant_id, ip_version=4,
                        cidr="172.16.0.0/24",
                        gateway_ip=neutron_attrs.ATTR_NOT_SPECIFIED,
                        dns_nameservers=neutron_attrs.ATTR_NOT_SPECIFIED,
                        host_routes=[
                            {"destination": "0.0.0.0/0",
                             "nexthop": "172.16.0.4"},
                            {"destination": "0.0.0.0/0",
                             "nexthop": "172.16.0.4"}],
                        allocation_pools=[{"start": "172.16.0.5",
                                           "end": "172.16.0.254"}],
                        enable_dhcp=None))
        with self._stubs(
            subnet=subnet["subnet"],
            routes=routes
        ) as (subnet_create, dns_create, route_create):
            dns_nameservers = subnet["subnet"].pop("dns_nameservers")
            gateway_ip = subnet["subnet"].pop("gateway_ip")
            subnet_request = copy.deepcopy(subnet)
            subnet_request["subnet"]["dns_nameservers"] = dns_nameservers
            subnet_request["subnet"]["gateway_ip"] = gateway_ip
            with self.assertRaises(q_exc.DuplicateRouteConflict):
                self.plugin.create_subnet(self.context, subnet_request)

    def test_create_subnet_default_route_gateway_ip(self):
        """Host_routes precedence

        If default route (host_routes) and gateway_ip are both provided,
        then host_route takes precedence.
        """

        routes = [dict(cidr="0.0.0.0/0", gateway="172.16.0.4")]
        subnet = dict(
            subnet=dict(network_id=1,
                        tenant_id=self.context.tenant_id, ip_version=4,
                        cidr="172.16.0.0/24",
                        gateway_ip="172.16.0.3",
                        dns_nameservers=neutron_attrs.ATTR_NOT_SPECIFIED,
                        host_routes=[{"destination": "0.0.0.0/0",
                                      "nexthop": "172.16.0.4"}],
                        allocation_pools=[{"start": "172.16.0.5",
                                           "end": "172.16.0.254"}],
                        enable_dhcp=None))
        with self._stubs(
            subnet=subnet["subnet"],
            routes=routes
        ) as (subnet_create, dns_create, route_create):
            dns_nameservers = subnet["subnet"].pop("dns_nameservers")
            subnet_request = copy.deepcopy(subnet)
            subnet_request["subnet"]["dns_nameservers"] = dns_nameservers
            res = self.plugin.create_subnet(self.context, subnet_request)
            self.assertEqual(subnet_create.call_count, 1)
            self.assertEqual(dns_create.call_count, 0)
            self.assertEqual(route_create.call_count, 1)
            self.assertEqual(res["gateway_ip"], "172.16.0.4")
            for key in subnet["subnet"].keys():
                if key == "gateway_ip":
                    self.assertEqual(res[key], "172.16.0.4")
                elif key == "host_routes":
                    self.assertEqual(len(res[key]), 0)
                else:
                    self.assertEqual(res[key], subnet["subnet"][key])

    def test_create_subnet_null_gateway_no_routes(self):
        """A subnet with a NULL gateway IP shouldn't create routes."""

        routes = [dict(cidr="0.0.0.0/0", gateway="172.16.0.4")]
        subnet = dict(
            subnet=dict(network_id=1,
                        tenant_id=self.context.tenant_id, ip_version=4,
                        cidr="172.16.0.0/24",
                        gateway_ip=None,
                        dns_nameservers=neutron_attrs.ATTR_NOT_SPECIFIED,
                        enable_dhcp=None))
        with self._stubs(
            subnet=subnet["subnet"],
            routes=routes
        ) as (subnet_create, dns_create, route_create):
            dns_nameservers = subnet["subnet"].pop("dns_nameservers")
            subnet_request = copy.deepcopy(subnet)
            subnet_request["subnet"]["dns_nameservers"] = dns_nameservers
            res = self.plugin.create_subnet(self.context, subnet_request)
            self.assertEqual(subnet_create.call_count, 1)
            self.assertEqual(dns_create.call_count, 0)
            self.assertEqual(route_create.call_count, 0)
            for key in subnet["subnet"].keys():
                if key == "gateway_ip":
                    self.assertIsNone(res[key])
                else:
                    self.assertEqual(res[key], subnet["subnet"][key])

    def test_create_subnet_routes_quota_pass(self):
        routes = (("0.0.0.0/0", "127.0.0.1"),
                  ("1.0.0.0/0", "127.0.0.1"),
                  ("2.0.0.0/0", "127.0.0.1"))
        host_routes = [{"destination": x, "nexthop": y} for x, y in routes]
        stub_routes = [{"cidr": x, "gateway": y} for x, y in routes]
        subnet = {"subnet":
                  {"cidr": "192.167.10.0/24", "created_at": datetime.now(),
                   "host_routes": host_routes, "id": 1, "ip_version": 4,
                   "network_id": 1, "tenant_id": self.context.tenant_id}}
        with self._stubs(subnet=subnet.get("subnet"), routes=stub_routes):
            self.plugin.create_subnet(self.context, subnet)

    def test_create_subnet_routes_quota_fail(self):
        routes = (("0.0.0.0/0", "127.0.0.1"),
                  ("1.0.0.0/0", "127.0.0.1"),
                  ("2.0.0.0/0", "127.0.0.1"),
                  ("3.0.0.0/0", "127.0.0.1"))
        host_routes = [{"destination": x, "nexthop": y} for x, y in routes]
        stub_routes = [{"cidr": x, "gateway": y} for x, y in routes]
        subnet = {"subnet":
                  {"cidr": "192.167.10.0/24", "created_at": datetime.now(),
                   "host_routes": host_routes, "id": 1, "ip_version": 4,
                   "network_id": 1, "tenant_id": self.context.tenant_id}}
        with self._stubs(subnet=subnet.get("subnet"), routes=stub_routes):
            with self.assertRaises(n_exc.OverQuota):
                self.plugin.create_subnet(self.context, subnet)

    def test_create_subnet_dns_quota_pass(self):
        nameservers = ["7.0.0.1", "7.0.0.2"]
        subnet = {"subnet":
                  {"cidr": "192.167.10.0/24", "created_at": datetime.now(),
                   "dns_nameservers": nameservers, "id": 1, "ip_version": 4,
                   "network_id": 1, "tenant_id": self.context.tenant_id}}
        with self._stubs(subnet=subnet.get("subnet"), dns=nameservers):
            self.plugin.create_subnet(self.context, subnet)

    def test_create_subnet_dns_quota_fail(self):
        nameservers = ["7.0.0.1", "7.0.0.2", "7.0.0.3"]
        subnet = {"subnet":
                  {"cidr": "192.167.10.0/24", "created_at": datetime.now(),
                   "dns_nameservers": nameservers, "id": 1, "ip_version": 4,
                   "network_id": 1, "tenant_id": self.context.tenant_id}}
        with self._stubs(subnet=subnet.get("subnet"), dns=nameservers):
            with self.assertRaises(n_exc.OverQuota):
                self.plugin.create_subnet(self.context, subnet)


class TestQuarkAllocationPoolCache(test_quark_plugin.TestQuarkPlugin):

    @contextlib.contextmanager
    def _stubs(self, has_subnet=True, host_routes=None, new_routes=None,
               find_routes=True, new_dns_servers=None, new_ip_policy=None,
               ip_version=4):
        if host_routes is None:
            host_routes = []
        if new_routes:
            new_routes = [models.Route(cidr=r["destination"],
                                       gateway=r["nexthop"],
                                       subnet_id=1)
                          for r in new_routes]
        if new_dns_servers:
            new_dns_servers = [models.DNSNameserver(
                ip=ip,
                subnet_id=1) for ip in new_dns_servers]
        if new_ip_policy:
            exc = [models.IPPolicyCIDR(cidr=excluded_cidr)
                   for excluded_cidr in new_ip_policy]
            new_ip_policy = models.IPPolicy(exclude=exc)

        if ip_version == 4:
            cidr = "172.16.0.0/24"
        else:
            cidr = "2607:f0d0:1002:51::0/64"

        subnet_mod = None
        if has_subnet:
            subnet = dict(
                id=0,
                network_id=1,
                tenant_id=self.context.tenant_id,
                ip_version=ip_version,
                cidr=cidr,
                host_routes=host_routes,
                dns_nameservers=["4.2.2.1", "4.2.2.2"],
                enable_dhcp=None,
                _allocation_pool_cache=None)

            dns_ips = subnet.pop("dns_nameservers", [])
            host_routes = subnet.pop("host_routes", [])
            exclude = [models.IPPolicyCIDR(cidr="172.16.0.0/32"),
                       models.IPPolicyCIDR(cidr="172.16.0.255/32")]
            subnet_mod = models.Subnet(
                ip_policy=models.IPPolicy(exclude=exclude),
                network=models.Network(id=1)
            )
            subnet_mod.update(subnet)

            subnet_mod["dns_nameservers"] = [models.DNSNameserver(ip=ip)
                                             for ip in dns_ips]
            subnet_mod["routes"] = [models.Route(cidr=r["destination"],
                                                 gateway=r["nexthop"],
                                                 subnet_id=subnet_mod["id"])
                                    for r in host_routes]
        with contextlib.nested(
            mock.patch("quark.db.api.subnet_find"),
            mock.patch("quark.db.api.subnet_update"),
            mock.patch("quark.db.api.dns_create"),
            mock.patch("quark.db.api.route_find"),
            mock.patch("quark.db.api.route_update"),
            mock.patch("quark.db.api.route_create"),
        ) as (subnet_find, subnet_update,
              dns_create, route_find, route_update, route_create):
            subnet_find.return_value = subnet_mod
            if has_subnet:
                route_find.return_value = (subnet_mod["routes"][0] if
                                           subnet_mod["routes"] and
                                           find_routes else None)
                new_subnet_mod = models.Subnet()
                new_subnet_mod.update(subnet_mod)
                new_subnet_mod.update(dict(id=1))
                if new_routes:
                    new_subnet_mod["routes"] = new_routes
                if new_dns_servers:
                    new_subnet_mod["dns_nameservers"] = new_dns_servers
                if new_ip_policy:
                    new_subnet_mod["ip_policy"] = new_ip_policy
                subnet_update.return_value = new_subnet_mod
            yield subnet_mod

    @mock.patch("quark.db.api.subnet_update_set_alloc_pool_cache")
    def test_update_subnet_allocation_pools_invalidate_cache(self, set_cache):
        og = cfg.CONF.QUARK.allow_allocation_pool_update
        cfg.CONF.set_override('allow_allocation_pool_update', True, 'QUARK')
        with self._stubs() as subnet_found:
            pools = [dict(start="172.16.0.1", end="172.16.0.12")]
            s = dict(subnet=dict(allocation_pools=pools))
            self.plugin.update_subnet(self.context, 1, s)
            self.assertEqual(set_cache.call_count, 1)
            set_cache.assert_called_with(self.context, subnet_found)
        cfg.CONF.set_override('allow_allocation_pool_update', og, 'QUARK')

    @mock.patch("quark.db.api.subnet_update_set_alloc_pool_cache")
    def test_get_subnet_set_alloc_cache_if_cache_is_none(self, set_cache):
        with self._stubs() as subnet_found:
            self.plugin.get_subnet(self.context, 1)
            self.assertEqual(set_cache.call_count, 1)
            set_cache.assert_called_with(self.context, subnet_found,
                                         [dict(start="172.16.0.1",
                                               end="172.16.0.254")])


class TestQuarkUpdateSubnet(test_quark_plugin.TestQuarkPlugin):
    DEFAULT_ROUTE = [dict(destination="0.0.0.0/0",
                          nexthop="172.16.0.1")]

    @contextlib.contextmanager
    def _stubs(self, has_subnet=True, host_routes=None, new_routes=None,
               find_routes=True, new_dns_servers=None, new_ip_policy=None,
               ip_version=4):
        if host_routes is None:
            host_routes = []
        if new_routes:
            new_routes = [models.Route(cidr=r["destination"],
                                       gateway=r["nexthop"],
                                       subnet_id=1)
                          for r in new_routes]
        if new_dns_servers:
            new_dns_servers = [models.DNSNameserver(
                ip=ip,
                subnet_id=1) for ip in new_dns_servers]
        if new_ip_policy:
            exc = [models.IPPolicyCIDR(cidr=excluded_cidr)
                   for excluded_cidr in new_ip_policy]
            new_ip_policy = models.IPPolicy(exclude=exc)

        if ip_version == 4:
            cidr = "172.16.0.0/24"
        else:
            cidr = "2607:f0d0:1002:51::0/64"

        subnet_mod = None
        if has_subnet:
            subnet = dict(
                id=0,
                network_id=1,
                tenant_id=self.context.tenant_id,
                ip_version=ip_version,
                cidr=cidr,
                host_routes=host_routes,
                dns_nameservers=["4.2.2.1", "4.2.2.2"],
                enable_dhcp=None)

            dns_ips = subnet.pop("dns_nameservers", [])
            host_routes = subnet.pop("host_routes", [])
            exclude = [models.IPPolicyCIDR(cidr="172.16.0.0/32"),
                       models.IPPolicyCIDR(cidr="172.16.0.255/32")]
            subnet_mod = models.Subnet(
                ip_policy=models.IPPolicy(exclude=exclude),
                network=models.Network(id=1)
            )
            subnet_mod.update(subnet)

            subnet_mod["dns_nameservers"] = [models.DNSNameserver(ip=ip)
                                             for ip in dns_ips]
            subnet_mod["routes"] = [models.Route(cidr=r["destination"],
                                                 gateway=r["nexthop"],
                                                 subnet_id=subnet_mod["id"])
                                    for r in host_routes]
        with contextlib.nested(
            mock.patch("quark.db.api.subnet_find"),
            mock.patch("quark.db.api.subnet_update"),
            mock.patch("quark.db.api.dns_create"),
            mock.patch("quark.db.api.route_find"),
            mock.patch("quark.db.api.route_update"),
            mock.patch("quark.db.api.route_create"),
        ) as (subnet_find, subnet_update,
              dns_create,
              route_find, route_update, route_create):
            subnet_find.return_value = subnet_mod
            if has_subnet:
                route_find.return_value = (subnet_mod["routes"][0] if
                                           subnet_mod["routes"] and
                                           find_routes else None)
                new_subnet_mod = models.Subnet()
                new_subnet_mod.update(subnet_mod)
                new_subnet_mod.update(dict(id=1))
                if new_routes:
                    new_subnet_mod["routes"] = new_routes
                if new_dns_servers:
                    new_subnet_mod["dns_nameservers"] = new_dns_servers
                if new_ip_policy:
                    new_subnet_mod["ip_policy"] = new_ip_policy
                subnet_update.return_value = new_subnet_mod
            yield dns_create, route_update, route_create

    def test_update_subnet_not_found(self):
        with self._stubs(has_subnet=False):
            with self.assertRaises(n_exc.SubnetNotFound):
                self.plugin.update_subnet(self.context, 1, {})

    def test_update_subnet_dns_nameservers(self):
        new_dns_servers = ["1.1.1.2"]
        with self._stubs(
            host_routes=self.DEFAULT_ROUTE,
            new_dns_servers=new_dns_servers
        ) as (dns_create, route_update, route_create):
            req = dict(subnet=dict(dns_nameservers=new_dns_servers))
            res = self.plugin.update_subnet(self.context,
                                            1,
                                            req)
            self.assertEqual(dns_create.call_count, 1)
            self.assertEqual(route_create.call_count, 0)
            self.assertEqual(res["dns_nameservers"], new_dns_servers)

    def test_update_subnet_routes(self):
        new_routes = [dict(destination="10.0.0.0/24",
                           nexthop="1.1.1.1")]
        with self._stubs(
            host_routes=self.DEFAULT_ROUTE,
            new_routes=new_routes
        ) as (dns_create, route_update, route_create):
            req = dict(subnet=dict(
                host_routes=new_routes))
            res = self.plugin.update_subnet(self.context, 1, req)
            self.assertEqual(dns_create.call_count, 0)
            self.assertEqual(route_create.call_count, 1)
            self.assertEqual(len(res["host_routes"]), 1)
            self.assertEqual(res["host_routes"][0]["destination"],
                             "10.0.0.0/24")
            self.assertEqual(res["host_routes"][0]["nexthop"],
                             "1.1.1.1")
            self.assertIsNone(res["gateway_ip"])

    def test_update_subnet_gateway_ip_with_default_route_in_db(self):
        with self._stubs(
            host_routes=self.DEFAULT_ROUTE,
            new_routes=[dict(destination="0.0.0.0/0", nexthop="1.2.3.4")]
        ) as (dns_create, route_update, route_create):
            req = dict(subnet=dict(gateway_ip="1.2.3.4"))
            res = self.plugin.update_subnet(self.context, 1, req)
            self.assertEqual(dns_create.call_count, 0)
            self.assertEqual(route_create.call_count, 0)
            self.assertEqual(route_update.call_count, 1)
            self.assertEqual(len(res["host_routes"]), 0)
            self.assertEqual(res["gateway_ip"], "1.2.3.4")

    def test_update_subnet_gateway_ip_with_non_default_route_in_db(self):
        with self._stubs(
            host_routes=[dict(destination="1.1.1.1/8", nexthop="9.9.9.9")],
            find_routes=False,
            new_routes=[dict(destination="1.1.1.1/8", nexthop="9.9.9.9"),
                        dict(destination="0.0.0.0/0", nexthop="1.2.3.4")]
        ) as (dns_create, route_update, route_create):
            req = dict(subnet=dict(gateway_ip="1.2.3.4"))
            res = self.plugin.update_subnet(self.context, 1, req)
            self.assertEqual(dns_create.call_count, 0)
            self.assertEqual(route_create.call_count, 1)

            self.assertEqual(res["gateway_ip"], "1.2.3.4")

            self.assertEqual(len(res["host_routes"]), 1)
            res_tuples = [(r["destination"], r["nexthop"])
                          for r in res["host_routes"]]
            self.assertIn(("1.1.1.1/8", "9.9.9.9"), res_tuples)

    def test_update_subnet_gateway_ip_without_default_route_in_db(self):
        with self._stubs(
            host_routes=None,
            new_routes=[dict(destination="0.0.0.0/0", nexthop="1.2.3.4")]
        ) as (dns_create, route_update, route_create):
            req = dict(subnet=dict(gateway_ip="1.2.3.4"))
            res = self.plugin.update_subnet(self.context, 1, req)
            self.assertEqual(dns_create.call_count, 0)
            self.assertEqual(route_create.call_count, 1)
            self.assertEqual(len(res["host_routes"]), 0)
            self.assertEqual(res["gateway_ip"], "1.2.3.4")

    def test_update_subnet_gateway_ip_with_default_route_in_args(self):
        new_routes = [dict(destination="0.0.0.0/0",
                           nexthop="4.3.2.1")]
        with self._stubs(
            host_routes=self.DEFAULT_ROUTE,
            new_routes=new_routes
        ) as (dns_create, route_update, route_create):
            req = dict(subnet=dict(
                host_routes=new_routes,
                gateway_ip="1.2.3.4"))
            res = self.plugin.update_subnet(self.context, 1, req)
            self.assertEqual(dns_create.call_count, 0)
            self.assertEqual(route_create.call_count, 1)
            self.assertEqual(len(res["host_routes"]), 0)
            self.assertEqual(res["gateway_ip"], "4.3.2.1")

    def test_update_subnet_allocation_pools_invalid_outside(self):
        og = cfg.CONF.QUARK.allow_allocation_pool_update
        cfg.CONF.set_override('allow_allocation_pool_update', True, 'QUARK')
        og1 = cfg.CONF.QUARK.allow_allocation_pool_growth
        cfg.CONF.set_override('allow_allocation_pool_growth', True, 'QUARK')
        pools = [dict(start="172.16.1.10", end="172.16.1.20")]
        s = dict(subnet=dict(allocation_pools=pools))
        try:
            with self._stubs() as (dns_create, route_update, route_create):
                with self.assertRaises(n_exc_ext.OutOfBoundsAllocationPool):
                    self.plugin.update_subnet(self.context, 1, s)
        finally:
            cfg.CONF.set_override('allow_allocation_pool_update', og, 'QUARK')
            cfg.CONF.set_override('allow_allocation_pool_growth', og1, 'QUARK')

    def test_update_subnet_allocation_pools_zero(self):
        with self._stubs() as (dns_create, route_update, route_create):
            resp = self.plugin.update_subnet(self.context, 1,
                                             dict(subnet=dict()))
            self.assertEqual(resp["allocation_pools"],
                             [dict(start="172.16.0.1", end="172.16.0.254")])

    def test_update_subnet_allocation_pools_one(self):
        og = cfg.CONF.QUARK.allow_allocation_pool_update
        cfg.CONF.set_override('allow_allocation_pool_update', True, 'QUARK')
        pools = [dict(start="172.16.0.10", end="172.16.0.20")]
        s = dict(subnet=dict(allocation_pools=pools))
        with self._stubs(
            new_ip_policy=[
                '172.16.0.0/29', '172.16.0.8/31', '172.16.0.21/32',
                '172.16.0.22/31', '172.16.0.24/29', '172.16.0.32/27',
                '172.16.0.64/26', '172.16.0.128/25']
        ) as (dns_create, route_update, route_create):
            resp = self.plugin.update_subnet(self.context, 1, s)
            self.assertEqual(resp["allocation_pools"], pools)
        cfg.CONF.set_override('allow_allocation_pool_update', og, 'QUARK')

    def test_update_subnet_allocation_pools_two(self):
        og = cfg.CONF.QUARK.allow_allocation_pool_update
        cfg.CONF.set_override('allow_allocation_pool_update', True, 'QUARK')
        pools = [dict(start="172.16.0.10", end="172.16.0.20"),
                 dict(start="172.16.0.40", end="172.16.0.50")]
        s = dict(subnet=dict(allocation_pools=pools))
        with self._stubs(
            new_ip_policy=[
                '172.16.0.0/29', '172.16.0.8/31', '172.16.0.21/32',
                '172.16.0.22/31', '172.16.0.24/29', '172.16.0.32/29',
                '172.16.0.51/32', '172.16.0.52/30', '172.16.0.56/29',
                '172.16.0.64/26', '172.16.0.128/25']
        ) as (dns_create, route_update, route_create):
            resp = self.plugin.update_subnet(self.context, 1, s)
            self.assertEqual(resp["allocation_pools"], pools)
        cfg.CONF.set_override('allow_allocation_pool_update', og, 'QUARK')

    def test_update_subnet_allocation_pools_three(self):
        og = cfg.CONF.QUARK.allow_allocation_pool_update
        cfg.CONF.set_override('allow_allocation_pool_update', True, 'QUARK')
        pools = [dict(start="172.16.0.5", end="172.16.0.254")]
        s = dict(subnet=dict(allocation_pools=pools))
        with self._stubs(
            new_ip_policy=['172.16.0.0/30', '172.16.0.4/32', '172.16.0.255/32']
        ) as (dns_create, route_update, route_create):
            resp = self.plugin.update_subnet(self.context, 1, s)
            self.assertEqual(resp["allocation_pools"], pools)
        cfg.CONF.set_override('allow_allocation_pool_update', og, 'QUARK')

    def test_update_subnet_allocation_pools_four(self):
        og = cfg.CONF.QUARK.allow_allocation_pool_update
        cfg.CONF.set_override('allow_allocation_pool_update', True, 'QUARK')
        pools = [dict(start="2607:f0d0:1002:51::a",
                      end="2607:f0d0:1002:51:ffff:ffff:ffff:fffe")]
        s = dict(subnet=dict(allocation_pools=pools))
        with self._stubs(
            ip_version=6,
            new_ip_policy=[
                '2607:f0d0:1002:51::/125', '2607:f0d0:1002:51::8/127',
                '2607:f0d0:1002:51:ffff:ffff:ffff:ffff/128']
        ) as (dns_create, route_update, route_create):
            resp = self.plugin.update_subnet(self.context, 1, s)
            self.assertEqual(resp["allocation_pools"], pools)
        cfg.CONF.set_override('allow_allocation_pool_update', og, 'QUARK')

    def test_update_subnet_allocation_pools_invalid(self):
        og = cfg.CONF.QUARK.allow_allocation_pool_update
        cfg.CONF.set_override('allow_allocation_pool_update', False, 'QUARK')
        pools = [dict(start="172.16.0.1", end="172.16.0.250")]
        s = dict(subnet=dict(allocation_pools=pools))
        with self._stubs() as (dns_create, route_update, route_create):
            with self.assertRaises(n_exc.BadRequest):
                self.plugin.update_subnet(self.context, 1, s)
        cfg.CONF.set_override('allow_allocation_pool_update', og, 'QUARK')

    def test_update_subnet_conflicting_gateway(self):
        og = cfg.CONF.QUARK.allow_allocation_pool_update
        cfg.CONF.set_override('allow_allocation_pool_update', True, 'QUARK')
        pools = [dict(start="172.16.0.1", end="172.16.0.254")]
        s = dict(subnet=dict(allocation_pools=pools, gateway_ip="172.16.0.1"))
        with self._stubs(
            new_ip_policy=['172.16.0.0/30', '172.16.0.4/32', '172.16.0.255/32']
        ) as (dns_create, route_update, route_create):
            with self.assertRaises(
                    n_exc_ext.GatewayConflictWithAllocationPools):
                self.plugin.update_subnet(self.context, 1, s)
        cfg.CONF.set_override('allow_allocation_pool_update', og, 'QUARK')


class TestQuarkDeleteSubnet(test_quark_plugin.TestQuarkPlugin):
    def setUp(self):
        super(TestQuarkDeleteSubnet, self).setUp()
        self.context.is_admin = True

    @contextlib.contextmanager
    def _stubs(self, subnet, ips):
        ip_mods = []
        subnet_mod = None
        if subnet:
            subnet_mod = models.Subnet()
            subnet_mod.update(subnet)
        for ip in ips:
            ip_mod = models.IPAddress()
            ip_mod.update(ip)
            ip_mods.append(ip_mod)

        strategy_prefix = "quark.network_strategy.JSONStrategy"
        with contextlib.nested(
            mock.patch("quark.db.api.subnet_find"),
            mock.patch("quark.db.api.subnet_delete"),
            mock.patch("neutron.common.rpc.get_notifier"),
            mock.patch("%s.is_provider_network" % strategy_prefix)
        ) as (sub_find, sub_delete, get_notifier, is_provider_network):
            if subnet_mod:
                subnet_mod.allocated_ips = ip_mods
            sub_find.return_value = subnet_mod
            is_provider_network.return_value = True
            yield sub_delete

    def test_delete_subnet(self):
        subnet = dict(id=1)
        with self._stubs(subnet=subnet, ips=[]) as sub_delete:
            self.plugin.delete_subnet(self.context, 1)
            self.assertTrue(sub_delete.called)

    def test_delete_subnet_on_provider_network_non_admin(self):
        """Attempts deleting a subnet on a provider network

        The subnet's tenant_id is a non-admin tenant and
        attempts to delete somebody else's subnet.
        Report NotFound because the tenant is not supposed to know
        about that subnet.
        """
        self.context.is_admin = False
        self.context.tenant_id = 5
        subnet = dict(id=1, tenant_id=1)
        with self._stubs(subnet=subnet, ips=[]):
            with self.assertRaises(n_exc.SubnetNotFound):
                self.plugin.delete_subnet(self.context, 1)

    def test_delete_own_subnet_on_provider_network_non_admin(self):
        """Attempts deleting a subnet on a provider network

        The subnet's tenant_id is a non-admin tenant and
        attempts deletion of own subnet.
        Disallow deletion of subnets on provider network.
        """
        self.context.is_admin = False
        self.context.tenant_id = 5
        subnet = dict(id=1, tenant_id=5)
        with self._stubs(subnet=subnet, ips=[]):
            with self.assertRaises(n_exc.NotAuthorized):
                self.plugin.delete_subnet(self.context, 1)

    def test_delete_subnet_no_subnet_fails(self):
        with self._stubs(subnet=None, ips=[]):
            with self.assertRaises(n_exc.SubnetNotFound):
                self.plugin.delete_subnet(self.context, 1)

    def test_delete_subnet_has_allocated_ips_fails(self):
        subnet = dict(id=1)
        with self._stubs(subnet=subnet, ips=[{}]):
            with self.assertRaises(n_exc.SubnetInUse):
                self.plugin.delete_subnet(self.context, 1)


class TestSubnetsQuotas(test_quark_plugin.TestQuarkPlugin):
    @contextlib.contextmanager
    def _stubs(self, subnet_values, deleted_at=None):
        self.context.session.begin = mock.MagicMock()

        subnets = list()
        for s in subnet_values:
            s["network"] = models.Network()
            s["network"]["created_at"] = s["created_at"]
            s["dns_nameservers"] = []
            s["_allocation_pool_cache"] = None
            subnet = models.Subnet(**s)
            subnets.append(subnet)
        with contextlib.nested(
            mock.patch("quark.plugin_modules.subnets.get_subnets"),
            mock.patch("quark.db.api.subnet_find"),
            mock.patch("quark.db.api.network_find"),
            mock.patch("quark.db.api.subnet_create"),
            mock.patch("quark.db.api.subnet_delete"),
            mock.patch("neutron.common.rpc.get_notifier"),
            mock.patch("oslo_utils.timeutils.utcnow"),
            mock.patch("quark.plugin_modules.subnets._validate_subnet_cidr")
        ) as (get_subnets, sub_find, net_find, sub_create, sub_del, notify,
              time_func, sub_validate):
            sub_create.return_value = subnets[0]
            sub_find.return_value = subnets[0]
            retsubs = []
            if len(subnets) > 1:
                retsubs = subnets[1:]
            get_subnets.return_value = retsubs
            time_func.return_value = deleted_at
            yield notify

    def test_create_subnet_v4_alongside_v6_quota_pass(self):
        original_4 = cfg.CONF.QUOTAS.quota_v4_subnets_per_network
        original_6 = cfg.CONF.QUOTAS.quota_v6_subnets_per_network
        s = [dict(network_id=1, cidr="192.167.10.0/24",
                  tenant_id=1, id=1, created_at="123"),
             dict(network_id=1, cidr="::0/24",
                  tenant_id=1, id=2, created_at="123")]
        with self._stubs(s):
            cfg.CONF.set_override('quota_v4_subnets_per_network', 1, "QUOTAS")
            cfg.CONF.set_override('quota_v6_subnets_per_network', 1, "QUOTAS")
            self.plugin.create_subnet(self.context, dict(subnet=s[0]))
            cfg.CONF.set_override('quota_v4_subnets_per_network', original_4,
                                  "QUOTAS")
            cfg.CONF.set_override('quota_v6_subnets_per_network', original_6,
                                  "QUOTAS")

    def test_create_subnet_v4_quota_pass(self):
        original_4 = cfg.CONF.QUOTAS.quota_v4_subnets_per_network
        s = [dict(network_id=1, cidr="192.167.10.0/24",
                  tenant_id=1, id=1, created_at="123")]
        with self._stubs(s):
            cfg.CONF.set_override('quota_v4_subnets_per_network', 1, "QUOTAS")
            self.plugin.create_subnet(self.context, dict(subnet=s[0]))
            cfg.CONF.set_override('quota_v4_subnets_per_network', original_4,
                                  "QUOTAS")

    def test_create_subnet_v6_quota_pass(self):
        original_6 = cfg.CONF.QUOTAS.quota_v6_subnets_per_network
        s = [dict(network_id=1, cidr="::0/24",
                  tenant_id=1, id=1, created_at="123")]
        with self._stubs(s):
            cfg.CONF.set_override('quota_v6_subnets_per_network', 1, "QUOTAS")
            self.plugin.create_subnet(self.context, dict(subnet=s[0]))
            cfg.CONF.set_override('quota_v6_subnets_per_network', original_6,
                                  "QUOTAS")

    def test_create_subnet_v4_quota_fail(self):
        original_4 = cfg.CONF.QUOTAS.quota_v4_subnets_per_network
        s = [dict(network_id=1, cidr="192.167.10.0/24",
                  tenant_id=1, id=1, created_at="123"),
             dict(network_id=1, cidr="192.168.10.0/24",
                  tenant_id=1, id=2, created_at="124")]
        with self._stubs(s):
            cfg.CONF.set_override('quota_v4_subnets_per_network', 1, "QUOTAS")
            with self.assertRaises(n_exc.OverQuota):
                self.plugin.create_subnet(self.context, dict(subnet=s[0]))
            cfg.CONF.set_override('quota_v4_subnets_per_network', original_4,
                                  "QUOTAS")

    def test_create_subnet_v6_quota_fail(self):
        original_6 = cfg.CONF.QUOTAS.quota_v6_subnets_per_network
        s = [dict(network_id=1, cidr="::0/24",
                  tenant_id=1, id=1, created_at="123"),
             dict(network_id=1, cidr="::1/24",
                  tenant_id=1, id=2, created_at="124")]
        with self._stubs(s):
            cfg.CONF.set_override('quota_v6_subnets_per_network', 1, "QUOTAS")
            with self.assertRaises(n_exc.OverQuota):
                self.plugin.create_subnet(self.context, dict(subnet=s[0]))
            cfg.CONF.set_override('quota_v6_subnets_per_network', original_6,
                                  "QUOTAS")

    def test_create_subnet_zero_quota_fail(self):
        original_4 = cfg.CONF.QUOTAS.quota_v4_subnets_per_network
        s = [dict(network_id=1, cidr="192.167.10.0/24",
                  tenant_id=1, id=1, created_at="123")]
        with self._stubs(s):
            cfg.CONF.set_override('quota_v4_subnets_per_network', 0, "QUOTAS")
            with self.assertRaises(n_exc.OverQuota):
                self.plugin.create_subnet(self.context, dict(subnet=s[0]))
            cfg.CONF.set_override('quota_v4_subnets_per_network', original_4,
                                  "QUOTAS")

    def test_create_subnet_negative_one_quota_pass(self):
        original_4 = cfg.CONF.QUOTAS.quota_v4_subnets_per_network
        s = [dict(network_id=1, cidr="192.167.10.0/24",
                  tenant_id=1, id=1, created_at="123")]
        with self._stubs(s):
            cfg.CONF.set_override('quota_v4_subnets_per_network', 0, "QUOTAS")
            with self.assertRaises(n_exc.OverQuota):
                self.plugin.create_subnet(self.context, dict(subnet=s[0]))
            cfg.CONF.set_override('quota_v4_subnets_per_network', -1, "QUOTAS")
            self.plugin.create_subnet(self.context, dict(subnet=s[0]))
            cfg.CONF.set_override('quota_v4_subnets_per_network', original_4,
                                  "QUOTAS")


class TestQuarkDiagnoseSubnets(test_quark_plugin.TestQuarkPlugin):
    @contextlib.contextmanager
    def _stubs(self, subnets=None, routes=None):
        if routes is None:
            routes = []
        route_models = []
        for route in routes:
            r = models.Route()
            r.update(route)
            route_models.append(r)

        if isinstance(subnets, list):
            subnet_models = []
            for subnet in subnets:
                s_dict = subnet.copy()
                s_dict["routes"] = route_models
                s = models.Subnet(network=models.Network())
                s.update(s_dict)
                subnet_models.append(s)
        elif subnets:
            mod = models.Subnet(network=models.Network())
            mod.update(subnets)
            mod["routes"] = route_models
            subnet_models = mod
        else:
            subnet_models = None

        with mock.patch("quark.db.api.subnet_find") as subnet_find:
            subnet_find.return_value = subnet_models
            yield

    def test_diagnose_subnet_with_wildcard_id_no_existing_subnets(self):
        with self._stubs(subnets=[], routes=[]):
            expected = {'subnets': []}
            actual = self.plugin.diagnose_subnet(self.context.elevated(), "*",
                                                 None)
            self.assertEqual(expected, actual)

    def test_diagnose_subnet_not_authorized(self):
        with self._stubs(subnets=[], routes=[]):
            with self.assertRaises(n_exc.NotAuthorized):
                self.plugin.diagnose_subnet(self.context, "*", None)

    def test_diagnose_subnet_with_wildcard_with_existing_subnets(self):
        subnet_id = str(uuid.uuid4())
        route = dict(id=1, cidr="0.0.0.0/0", gateway="192.168.0.1")

        subnet = dict(id=subnet_id, network_id=1, name=subnet_id,
                      tenant_id=self.context.tenant_id, ip_version=4,
                      cidr="192.168.0.0/24", gateway_ip="192.168.0.1",
                      dns_nameservers=[],
                      enable_dhcp=None)

        with self._stubs(subnets=[subnet], routes=[route]):
            actual = self.plugin.diagnose_subnet(self.context.elevated(), "*",
                                                 None)
            self.maxDiff = None
            self.assertEqual(subnet["id"], actual["subnets"][0]["id"])

    def test_diagnose_subnet_with_regular_id(self):
        subnet_id = "12345"
        route = dict(id=1, cidr="0.0.0.0/0", gateway="192.168.0.1")

        subnet = dict(id=subnet_id, network_id=1, name=subnet_id,
                      tenant_id=self.context.tenant_id, ip_version=4,
                      cidr="192.168.0.0/24", gateway_ip="192.168.0.1",
                      dns_nameservers=[],
                      enable_dhcp=None)

        with self._stubs(subnets=subnet, routes=[route]):
            actual = self.plugin.diagnose_subnet(self.context.elevated(),
                                                 subnet_id, None)
            self.assertEqual(subnet["id"], actual["subnets"]["id"])


class TestQuarkCreateSubnetAttrFilters(test_quark_plugin.TestQuarkPlugin):
    @contextlib.contextmanager
    def _stubs(self):
        with contextlib.nested(
            mock.patch("quark.db.api.subnet_create"),
            mock.patch("quark.db.api.network_find"),
            mock.patch("quark.db.api.dns_create"),
            mock.patch("quark.db.api.route_create"),
            mock.patch("quark.plugin_views._make_subnet_dict"),
            mock.patch("quark.db.api.subnet_find"),
            mock.patch("neutron.common.rpc.get_notifier")
        ) as (subnet_create, net_find, dns_create, route_create, sub_dict,
              subnet_find, get_notifier):
            route_create.return_value = models.Route()
            yield subnet_create, net_find

    def test_create_subnet(self):
        subnet = {"subnet": {
            "network_id": 1, "tenant_id": self.context.tenant_id,
            "ip_version": 4, "cidr": "172.16.0.0/24",
            "gateway_ip": "0.0.0.0",
            "dns_nameservers": neutron_attrs.ATTR_NOT_SPECIFIED,
            "host_routes": neutron_attrs.ATTR_NOT_SPECIFIED,
            "enable_dhcp": None, "first_ip": 0, "last_ip": 1,
            "next_auto_assign_ip": 10}}

        with self._stubs() as (subnet_create, net_find):
            subnet_create.return_value = models.Subnet(
                cidr=subnet["subnet"]["cidr"])
            self.plugin.create_subnet(self.context, subnet)
            self.assertEqual(subnet_create.call_count, 1)
            subnet_create.assert_called_once_with(
                self.context, network_id=subnet["subnet"]["network_id"],
                tenant_id=subnet["subnet"]["tenant_id"],
                cidr=subnet["subnet"]["cidr"], network=net_find())

    def test_create_subnet_admin(self):
        subnet = {"subnet": {
            "network_id": 1, "tenant_id": self.context.tenant_id,
            "ip_version": 4, "cidr": "172.16.0.0/24",
            "gateway_ip": "0.0.0.0",
            "dns_nameservers": neutron_attrs.ATTR_NOT_SPECIFIED,
            "host_routes": neutron_attrs.ATTR_NOT_SPECIFIED,
            "enable_dhcp": None, "first_ip": 0, "last_ip": 1,
            "next_auto_assign_ip": 10}}

        admin_ctx = self.context.elevated()
        with self._stubs() as (subnet_create, net_find):
            subnet_create.return_value = models.Subnet(
                cidr=subnet["subnet"]["cidr"])
            self.plugin.create_subnet(admin_ctx, subnet)
            self.assertEqual(subnet_create.call_count, 1)
            subnet_create.assert_called_once_with(
                admin_ctx, network_id=subnet["subnet"]["network_id"],
                tenant_id=subnet["subnet"]["tenant_id"],
                cidr=subnet["subnet"]["cidr"], network=net_find(),
                next_auto_assign_ip=subnet["subnet"]["next_auto_assign_ip"])


class TestQuarkUpdateSubnetAttrFilters(test_quark_plugin.TestQuarkPlugin):
    @contextlib.contextmanager
    def _stubs(self):
        pool_mod = "quark.allocation_pool.AllocationPools"
        with contextlib.nested(
            mock.patch("quark.db.api.subnet_find"),
            mock.patch("quark.db.api.subnet_update"),
            mock.patch("quark.db.api.dns_create"),
            mock.patch("quark.db.api.route_find"),
            mock.patch("quark.db.api.route_update"),
            mock.patch("quark.db.api.route_create"),
            mock.patch(pool_mod),
            mock.patch("quark.plugin_views._make_subnet_dict")
        ) as (subnet_find, subnet_update, dns_create, route_find,
              route_update, route_create, make_subnet, gateway_exclude):
            yield subnet_update, subnet_find

    def test_update_subnet_attr_filters(self):
        subnet = {"subnet": {
            "network_id": 1, "tenant_id": self.context.tenant_id,
            "ip_version": 4, "cidr": "172.16.0.0/24",
            "gateway_ip": "0.0.0.0",
            "dns_nameservers": neutron_attrs.ATTR_NOT_SPECIFIED,
            "host_routes": neutron_attrs.ATTR_NOT_SPECIFIED,
            "enable_dhcp": None, "first_ip": 0, "last_ip": 1,
            "next_auto_assign_ip": 10, "do_not_use": False}}

        with self._stubs() as (subnet_update, subnet_find):
            self.plugin.update_subnet(self.context, 1, subnet)

            # NOTE(mdietz): the assertion here shows that, without admin,
            #               all of the attributes passed above are stripped
            #               from the request body. Otherwise, the attributes
            #               above would be passed as keyword arguments to the
            #               subnet_update db api call.
            subnet_update.assert_called_once_with(
                self.context, subnet_find())

    def test_update_subnet_attr_filters_admin(self):
        subnet = {"subnet": {
            "network_id": 1, "tenant_id": self.context.tenant_id,
            "ip_version": 4, "cidr": "172.16.0.0/24",
            "gateway_ip": "0.0.0.0",
            "dns_nameservers": neutron_attrs.ATTR_NOT_SPECIFIED,
            "host_routes": neutron_attrs.ATTR_NOT_SPECIFIED,
            "enable_dhcp": False, "first_ip": 0, "last_ip": 1,
            "next_auto_assign_ip": 10, "do_not_use": True}}

        admin_ctx = self.context.elevated()
        with self._stubs() as (subnet_update, subnet_find):
            self.plugin.update_subnet(admin_ctx, 1, subnet)
            subnet_update.assert_called_once_with(
                admin_ctx, subnet_find(),
                next_auto_assign_ip=subnet["subnet"]["next_auto_assign_ip"],
                tenant_id=subnet["subnet"]["tenant_id"],
                enable_dhcp=subnet["subnet"]["enable_dhcp"],
                do_not_use=subnet["subnet"]["do_not_use"])


class TestQuarkGetSubnetsShared(test_quark_plugin.TestQuarkPlugin):
    def setUp(self):
        super(TestQuarkGetSubnetsShared, self).setUp()
        self.strategy = {"public_network":
                         {"bridge": "xenbr0",
                          "subnets": {"4": "public_v4",
                                      "6": "public_v6"}}}
        self.strategy_json = json.dumps(self.strategy)
        self.old = plugin_views.STRATEGY
        plugin_views.STRATEGY = network_strategy.JSONStrategy(
            self.strategy_json)
        cfg.CONF.set_override("default_net_strategy", self.strategy_json,
                              "QUARK")

    def tearDown(self):
        plugin_views.STRATEGY = self.old

    @contextlib.contextmanager
    def _stubs(self, subnets=None):
        subnet_mods = []

        if isinstance(subnets, list):
            for sub in subnets:
                subnet_mod = models.Subnet()
                subnet_mod.update(sub)
                subnet_mods.append(subnet_mod)

        db_mod = "quark.db.api"
        db_api.STRATEGY = network_strategy.JSONStrategy(self.strategy_json)
        network_strategy.STRATEGY = network_strategy.JSONStrategy(
            self.strategy_json)

        with mock.patch("%s._subnet_find" % db_mod) as subnet_find:
            subnet_find.return_value = subnet_mods
            yield subnet_find

    def test_get_subnets_shared(self):
        sub0 = dict(id='public_v4', tenant_id="provider", name="public_v4",
                    _cidr="0.0.0.0/0", network_id="public_network")
        sub1 = dict(id='public_v6', tenant_id="provider", name="public_v6",
                    _cidr="::/0", network_id="public_network")

        with self._stubs(subnets=[sub0, sub1]) as subnet_find:
            ret = self.plugin.get_subnets(self.context, None, None, None,
                                          False, {"shared": [True]})
            for sub in ret:
                self.assertEqual("public_network", sub["network_id"])

            subnet_find.assert_called_with(self.context, None, None, False,
                                           None, None,
                                           join_routes=True,
                                           defaults=["public_v4", "public_v6"],
                                           join_dns=True,
                                           join_pool=True,
                                           provider_query=False)

    def test_get_subnets_shared_false(self):
        sub0 = dict(id='public_v4', tenant_id="provider", name="public_v4",
                    _cidr="0.0.0.0/0", network_id="public_network")
        sub1 = dict(id='public_v6', tenant_id="provider", name="public_v6",
                    _cidr="::/0", network_id="public_network")

        with self._stubs(subnets=[sub0, sub1]) as subnet_find:
            self.plugin.get_subnets(self.context, None, None, None,
                                    False, {"shared": [False]})
            invert = db_api.INVERT_DEFAULTS
            subnet_find.assert_called_with(self.context, None, None, False,
                                           None, None,
                                           defaults=[invert, "public_v4",
                                                     "public_v6"],
                                           provider_query=False,
                                           join_routes=True, join_dns=True,
                                           join_pool=True)

    def test_get_subnets_no_shared(self):
        sub0 = dict(id='public_v4', tenant_id="provider", name="public_v4",
                    _cidr="0.0.0.0/0", network_id="public_network")
        sub1 = dict(id='tenant_v4', tenant_id="tenant", name="tenant_v4",
                    _cidr="0.0.0.0/0", network_id="tenant_network")

        with self._stubs(subnets=[sub0, sub1]) as subnet_find:
            self.plugin.get_subnets(self.context, None, None, None,
                                    False)
            subnet_find.assert_called_with(self.context, None, None, False,
                                           None, None,
                                           defaults=[],
                                           provider_query=False,
                                           join_routes=True, join_dns=True,
                                           join_pool=True)



# Copyright (c) 2015 Rackspace Hosting Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from neutron._i18n import _
from oslo_config import cfg

CONF = cfg.CONF


class Capabilities(object):
    SECURITY_GROUPS = "security_groups"
    EGRESS = "egress"
    TENANT_NETWORK_SG = "tenant_network_sg"
    IP_BILLING = "ip_billing"


quark_opts = [
    cfg.ListOpt("environment_capabilities",
                default=",".join([Capabilities.SECURITY_GROUPS]),
                help=_("Capabilities supported by a given environment's"
                       "deployment.")),
]


CONF.register_opts(quark_opts, "QUARK")


class has_capability(object):
    def __init__(self, capability):
        self.capability = capability

    def __call__(self, f):
        def wrapped(*args, **kwargs):
            if self.capability in CONF.QUARK.environment_capabilities:
                return f(*args, **kwargs)
        return wrapped



"""Add async transactions table

Revision ID: 271cce54e15b
Revises: 2a116b962c95
Create Date: 2016-06-15 09:24:29.941684

"""

# revision identifiers, used by Alembic.
revision = '271cce54e15b'
down_revision = '2a116b962c95'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.create_table('quark_async_transactions',
                    sa.Column('created_at', sa.DateTime(), nullable=True),
                    sa.Column('id', sa.String(36), nullable=False),
                    sa.Column('tenant_id', sa.String(255), nullable=False),
                    sa.Column('action', sa.String(255), nullable=False),
                    sa.Column('completed', sa.Boolean()),
                    sa.PrimaryKeyConstraint('id'),
                    mysql_engine='InnoDB')
    op.create_index(op.f('ix_quark_async_transactions_tenant_id'),
                    'quark_async_transactions', ['tenant_id'],
                    unique=False)


def downgrade():
    op.drop_table('quark_async_transactions')



# Copyright 2013 Rackspace Hosting Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
NVP client driver for Quark
"""

import contextlib
import random

import aiclib
from neutron.extensions import securitygroup as sg_ext
from oslo_config import cfg
from oslo_log import log as logging

from quark.drivers import base
from quark.drivers import security_groups as sg_driver
from quark.environment import Capabilities
from quark import exceptions as q_exc
from quark import segment_allocations
from quark import utils

LOG = logging.getLogger(__name__)

CONF = cfg.CONF

nvp_opts = [
    cfg.IntOpt('max_ports_per_switch',
               default=0,
               help=_('Maximum amount of NVP ports on an NVP lswitch')),
    cfg.StrOpt('default_tz_type',
               help=_('The type of connector to use for the default tz'),
               default="stt"),
    cfg.ListOpt('additional_default_tz_types',
                default=[],
                help=_('List of additional default tz types to bind to the '
                       'default tz')),
    cfg.StrOpt('default_tz',
               help=_('The default transport zone UUID')),
    cfg.ListOpt('controller_connection',
                default=[],
                help=_('NVP Controller connection string')),
    cfg.IntOpt('max_rules_per_group',
               default=30,
               help=_('Maxiumum size of NVP SecurityRule list per group')),
    cfg.IntOpt('max_rules_per_port',
               default=30,
               help=_('Maximum rules per NVP lport across all groups')),
    cfg.IntOpt('backoff',
               default=0,
               help=_('Base seconds for exponential backoff')),
    cfg.BoolOpt("random_initial_controller",
                default=False,
                help=_("Whether or not to use a random controller or the "
                       "first controller when neutron starts up")),
    cfg.IntOpt("connection_switching_threshold",
               default=0,
               help=_("Number of times to use a connection before forcing "
                      "the next connection in the list to be used. A value "
                      "of 0 means only switch on Exceptions.")),
    cfg.BoolOpt("connection_switching_random",
                default=False,
                help=_("Determines whether connections are switched randomly "
                       "or using the default round-robin.")),
    cfg.IntOpt("operation_retries",
               default=3,
               help=_("Number of times to attempt to perform operations in "
                      "NVP.")),
]

physical_net_type_map = {
    "stt": "stt",
    "gre": "gre",
    "flat": "bridge",
    "bridge": "bridge",
    "vlan": "bridge",
    "local": "local",
}

CONF.register_opts(nvp_opts, "NVP")
SA_REGISTRY = segment_allocations.REGISTRY


class TransportZoneBinding(object):

    net_type = None

    def add(self, context, switch, tz_id, network_id):
        raise NotImplementedError()

    def remove(self, context, tz_id, network_id):
        raise NotImplementedError()


class VXLanTransportZoneBinding(TransportZoneBinding):

    net_type = 'vxlan'

    def add(self, context, switch, tz_id, network_id):
        driver = SA_REGISTRY.get_strategy(self.net_type)
        alloc = driver.allocate(context, tz_id, network_id)
        switch.transport_zone(
            tz_id, self.net_type, vxlan_id=alloc["id"])

    def remove(self, context, tz_id, network_id):
        driver = SA_REGISTRY.get_strategy(self.net_type)
        driver.deallocate(context, tz_id, network_id)


# A map of net_type (vlan, vxlan) to TransportZoneBinding impl.
TZ_BINDINGS = {
    VXLanTransportZoneBinding.net_type: VXLanTransportZoneBinding()
}


def _tag_roll(tags):
    return [{'scope': k, 'tag': v} for k, v in tags]


def _tag_unroll(tags):
    return dict((t['scope'], t['tag']) for t in tags)


class NVPDriver(base.BaseDriver):
    def __init__(self):
        self.nvp_connections = []
        self.conn_index = 0
        self.limits = {'max_ports_per_switch': 0,
                       'max_rules_per_group': 0,
                       'max_rules_per_port': 0}
        self.sg_driver = None
        if Capabilities.SECURITY_GROUPS in CONF.QUARK.environment_capabilities:
            self.sg_driver = sg_driver.SecurityGroupDriver()

        super(NVPDriver, self).__init__()

    @classmethod
    def get_name(klass):
        return "NVP"

    def load_config(self):
        # NOTE(mdietz): What does default_tz actually mean?
        #               We don't have one default.
        # NOTE(jkoelker): Transport Zone
        # NOTE(mdietz): :-/ tz isn't the issue. default is
        default_tz = CONF.NVP.default_tz
        LOG.info("Loading NVP settings " + str(default_tz))
        connections = CONF.NVP.controller_connection
        backoff = CONF.NVP.backoff

        self.limits.update({
            'max_ports_per_switch': CONF.NVP.max_ports_per_switch,
            'max_rules_per_group': CONF.NVP.max_rules_per_group,
            'max_rules_per_port': CONF.NVP.max_rules_per_port})
        LOG.info("Loading NVP settings " + str(connections))

        for conn in connections:
            (ip, port, user, pw, req_timeout,
             http_timeout, retries, redirects) = conn.split(":")

            self.nvp_connections.append(dict(ip_address=ip,
                                        port=port,
                                        username=user,
                                        password=pw,
                                        req_timeout=req_timeout,
                                        http_timeout=int(http_timeout),
                                        retries=int(retries),
                                        redirects=redirects,
                                        default_tz=default_tz,
                                        backoff=backoff,
                                        usages=0))

        if connections:
            if CONF.NVP.random_initial_controller:
                self.conn_index = self._new_random_index(-1, len(connections))

            LOG.info("NVP Driver config loaded. Starting with controller %s" %
                     self.nvp_connections[self.conn_index]["ip_address"])
        else:
            LOG.critical("No NVP connection configurations found!")

    def _new_random_index(self, current, index_range):
        new_index = current
        while new_index == current:
            new_index = random.randint(0, index_range - 1)
        return new_index

    def _connection(self):
        if len(self.nvp_connections) == 0:
            raise q_exc.NoBackendConnectionsDefined(
                msg="No NVP connections defined cannot continue")

        conn = self.nvp_connections[self.conn_index]

        if CONF.NVP.connection_switching_threshold > 0:
            # NOTE(mdietz): This is racy. See get_connection below.
            conn["usages"] += 1
            if conn["usages"] >= CONF.NVP.connection_switching_threshold:
                conn["usages"] = 0
                self._next_connection()
                conn = self.nvp_connections[self.conn_index]

        if "connection" not in conn:
            scheme = conn["port"] == "443" and "https" or "http"
            uri = "%s://%s:%s" % (scheme, conn["ip_address"], conn["port"])
            user = conn['username']
            passwd = conn['password']
            timeout = conn['http_timeout']
            retries = conn['retries']
            backoff = conn['backoff']
            conn["connection"] = aiclib.nvp.Connection(uri,
                                                       username=user,
                                                       password=passwd,
                                                       timeout=timeout,
                                                       retries=retries,
                                                       backoff=backoff)
        return conn["connection"]

    def _next_connection(self):
        # TODO(anyone): Do we want to drop and create new connections at some
        #               point? What about recycling them after a certain
        #               number of usages or time, proactively?
        LOG.info("Switching NVP connections...")
        conn_len = len(self.nvp_connections)
        if conn_len and conn_len > 1:
            if CONF.NVP.connection_switching_random:
                self.conn_index = self._new_random_index(self.conn_index,
                                                         conn_len)
                LOG.info("New connection chosen at random is %s" %
                         self.nvp_connections[self.conn_index]["ip_address"])
            else:
                self.conn_index = (self.conn_index + 1) % conn_len
                LOG.info("New connection chosen round-robin is %s" %
                         self.nvp_connections[self.conn_index]["ip_address"])
        else:
            LOG.info("No other connections to choose from")

    @contextlib.contextmanager
    def get_connection(self):
        try:
            yield self._connection()
        except Exception:
            # This is racy. A pile-up of failures could occur on one
            # controller, causing them to rapidly switch and fail back
            # to the original failing controller. However, we can't be in
            # the business of implementing a load balancer inside the code.
            # TODO(anyone): Investigate whether NVP is now giving us sequence
            #               IDs. If so, rapid round-robining becomes plausible,
            #               though we don't have an easy path to dropping bad
            #               controllers.
            self._next_connection()
            raise

    def create_network(self, context, network_name, tags=None,
                       network_id=None, **kwargs):
        return self._lswitch_create(context, network_name, tags,
                                    network_id, **kwargs)

    def delete_network(self, context, network_id):
        lswitches = self._lswitches_for_network(context, network_id).results()
        for switch in lswitches["results"]:
            try:
                self._lswitch_delete(context, switch["uuid"])
                # NOTE(morgabra) If we haven't thrown here, we can be sure the
                # resource was deleted from NSX. So we give a chance for any
                # previously allocated segment ids to deallocate.
                self._remove_default_tz_bindings(
                    context, network_id)
            except aiclib.core.AICException as ae:
                if ae.code == 404:
                    LOG.info("LSwitch/Network %s not found in NVP."
                             " Ignoring explicitly. Code: %s, Message: %s"
                             % (network_id, ae.code, ae.message))
                else:
                    LOG.info("AICException deleting LSwitch/Network %s in NVP."
                             " Ignoring explicitly. Code: %s, Message: %s"
                             % (network_id, ae.code, ae.message))
            except Exception as e:
                LOG.info("Failed to delete LSwitch/Network %s in NVP."
                         " Ignoring explicitly. Message: %s"
                         % (network_id, e.args[0]))

    def _collect_lswitch_info(self, lswitch, get_status):
        info = {
            'port_isolation_enabled': lswitch['port_isolation_enabled'],
            'display_name': lswitch['display_name'],
            'uuid': lswitch['uuid'],
            'transport_zones': lswitch['transport_zones'],
        }
        info.update(_tag_unroll(lswitch['tags']))
        if get_status:
            status = lswitch.pop('_relations')['LogicalSwitchStatus']
            info.update({
                'lport_stats': {
                    'fabric_up': status['lport_fabric_up_count'],
                    'admin_up': status['lport_admin_up_count'],
                    'link_up': status['lport_link_up_count'],
                    'count': status['lport_count'],
                }, 'fabric_status': status['fabric_status'],
            })
        return info

    def diag_network(self, context, network_id, get_status, **kwargs):
        switches = self._lswitch_status_query(context, network_id)['results']
        return {'logical_switches': [self._collect_lswitch_info(s, get_status)
                for s in switches]}

    def create_port(self, context, network_id, port_id, status=True,
                    security_groups=None, device_id="", **kwargs):
        security_groups = security_groups or []
        tenant_id = context.tenant_id
        lswitch = self._create_or_choose_lswitch(context, network_id)

        @utils.retry_loop(CONF.NVP.operation_retries)
        def _create_lswitch_port():
            with self.get_connection() as connection:
                port = connection.lswitch_port(lswitch)
                port.admin_status_enabled(status)
                if not self.sg_driver:
                    nvp_group_ids = self._get_security_groups_for_port(
                        context, security_groups)
                    port.security_profiles(nvp_group_ids)
                tags = [dict(tag=network_id, scope="neutron_net_id"),
                        dict(tag=port_id, scope="neutron_port_id"),
                        dict(tag=tenant_id, scope="os_tid"),
                        dict(tag=device_id, scope="vm_id")]
                LOG.debug("Creating port on switch %s" % lswitch)
                port.tags(tags)
                res = port.create()
                try:
                    """Catching odd NVP returns here will make it safe to
                    assume that NVP returned something correct."""
                    res["lswitch"] = lswitch
                except TypeError:
                    LOG.exception("Unexpected return from NVP: %s" % res)
                    raise
                port = connection.lswitch_port(lswitch)
                port.uuid = res["uuid"]
                port.attachment_vif(port_id)
                return res
        return _create_lswitch_port()

    @utils.retry_loop(CONF.NVP.operation_retries)
    def update_port(self, context, port_id, mac_address=None, device_id=None,
                    status=True, security_groups=None, **kwargs):
        if not self.sg_driver:
            security_groups = security_groups or []
        else:
            kwargs.update({'security_groups': security_groups})
        with self.get_connection() as connection:
            if self.sg_driver:
                kwargs.update({'mac_address': mac_address,
                               'device_id': device_id})
                self.sg_driver.update_port(**kwargs)
            lswitch_id = self._lswitch_from_port(context, port_id)
            port = connection.lswitch_port(lswitch_id, port_id)
            if not self.sg_driver:
                nvp_group_ids = self._get_security_groups_for_port(
                    context, security_groups)
                if nvp_group_ids:
                    port.security_profiles(nvp_group_ids)
            port.admin_status_enabled(status)
            return port.update()

    @utils.retry_loop(CONF.NVP.operation_retries)
    def delete_port(self, context, port_id, **kwargs):
        with self.get_connection() as connection:
            lswitch_uuid = kwargs.get('lswitch_uuid', None)
            try:
                if not lswitch_uuid:
                    lswitch_uuid = self._lswitch_from_port(context, port_id)
                LOG.debug("Deleting port %s from lswitch %s"
                          % (port_id, lswitch_uuid))
                connection.lswitch_port(lswitch_uuid, port_id).delete()
                if self.sg_driver:
                    self.sg_driver.delete_port(**kwargs)
            except aiclib.core.AICException as ae:
                if ae.code == 404:
                    LOG.info("LSwitchPort/Port %s not found in NVP."
                             " Ignoring explicitly. Code: %s, Message: %s"
                             % (port_id, ae.code, ae.message))
                else:
                    LOG.info("AICException deleting LSwitchPort/Port %s in "
                             "NVP. Ignoring explicitly. Code: %s, Message: %s"
                             % (port_id, ae.code, ae.message))

            except Exception as e:
                LOG.info("Failed to delete LSwitchPort/Port %s in NVP."
                         " Ignoring explicitly. Message: %s"
                         % (port_id, e.args[0]))

    def _collect_lport_info(self, lport, get_status):
        info = {
            'mirror_targets': lport['mirror_targets'],
            'display_name': lport['display_name'],
            'portno': lport['portno'],
            'allowed_address_pairs': lport['allowed_address_pairs'],
            'nvp_security_groups': lport['security_profiles'],
            'uuid': lport['uuid'],
            'admin_status_enabled': lport['admin_status_enabled'],
            'queue_uuid': lport['queue_uuid'],
        }
        if get_status:
            stats = lport['statistics']
            status = lport['status']
            lswitch = {
                'uuid': status['lswitch']['uuid'],
                'display_name': status['lswitch']['display_name'],
            }
            lswitch.update(_tag_unroll(status['lswitch']['tags']))
            info.update({
                'statistics': {
                    'recieved': {
                        'packets': stats['rx_packets'],
                        'bytes': stats['rx_bytes'],
                        'errors': stats['rx_errors']},
                    'transmitted': {
                        'packets': stats['tx_packets'],
                        'bytes': stats['tx_bytes'],
                        'errors': stats['tx_errors']
                    },
                },
                'status': {
                    'link_status_up': status['link_status_up'],
                    'admin_status_up': status['admin_status_up'],
                    'fabric_status_up': status['fabric_status_up'],
                },
                'lswitch': lswitch,
            })
        info.update(_tag_unroll(lport['tags']))
        return info

    def diag_port(self, context, port_id, get_status=False, **kwargs):
        with self.get_connection() as connection:
            lswitch_uuid = self._lswitch_from_port(context, port_id)
            lswitch_port = connection.lswitch_port(lswitch_uuid, port_id)

            query = lswitch_port.query()
            query.relations("LogicalPortAttachment")
            results = query.results()
            if results['result_count'] == 0:
                return {'lport': "Logical port not found."}

            config = results['results'][0]
            relations = config.pop('_relations')
            config['attachment'] = relations['LogicalPortAttachment']['type']
            if get_status:
                config['status'] = lswitch_port.status()
                config['statistics'] = lswitch_port.statistics()
            return {'lport': self._collect_lport_info(config, get_status)}

    def _get_network_details(self, context, network_id, switches):
        name, phys_net, phys_type, segment_id = None, None, None, None
        for res in switches["results"]:
            name = res["display_name"]
            for zone in res["transport_zones"]:
                phys_net = zone["zone_uuid"]
                phys_type = zone["transport_type"]
                if "binding_config" in zone:
                    binding = zone["binding_config"]
                    segment_id = binding["vlan_translation"][0]["transport"]
                break
            return dict(network_name=name, phys_net=phys_net,
                        phys_type=phys_type, segment_id=segment_id)
        return {}

    def create_security_group(self, context, group_name, **group):
        tenant_id = context.tenant_id
        with self.get_connection() as connection:
            group_id = group.get('group_id')
            profile = connection.securityprofile()
            if group_name:
                profile.display_name(group_name)
            ingress_rules = group.get('port_ingress_rules', [])
            egress_rules = group.get('port_egress_rules', [])

            if (len(ingress_rules) + len(egress_rules) >
                    self.limits['max_rules_per_group']):
                raise q_exc.DriverLimitReached(limit="rules per group")

            if egress_rules:
                profile.port_egress_rules(egress_rules)
            if ingress_rules:
                profile.port_ingress_rules(ingress_rules)
            tags = [dict(tag=group_id, scope="neutron_group_id"),
                    dict(tag=tenant_id, scope="os_tid")]
            LOG.debug("Creating security profile %s" % group_name)
            profile.tags(tags)
            return profile.create()

    def delete_security_group(self, context, group_id, **kwargs):
        guuid = self._get_security_group_id(context, group_id)
        with self.get_connection() as connection:
            LOG.debug("Deleting security profile %s" % group_id)
            connection.securityprofile(guuid).delete()

    def update_security_group(self, context, group_id, **group):
        query = self._get_security_group(context, group_id)
        with self.get_connection() as connection:
            profile = connection.securityprofile(query.get('uuid'))

            ingress_rules = group.get('port_ingress_rules',
                                      query.get('logical_port_ingress_rules'))
            egress_rules = group.get('port_egress_rules',
                                     query.get('logical_port_egress_rules'))

            if (len(ingress_rules) + len(egress_rules) >
                    self.limits['max_rules_per_group']):
                raise q_exc.DriverLimitReached(limit="rules per group")

            if group.get('name', None):
                profile.display_name(group['name'])
            if group.get('port_ingress_rules', None) is not None:
                profile.port_ingress_rules(ingress_rules)
            if group.get('port_egress_rules', None) is not None:
                profile.port_egress_rules(egress_rules)
            return profile.update()

    def _update_security_group_rules(self, context, group_id, rule, operation,
                                     checks):
        groupd = self._get_security_group(context, group_id)
        direction, secrule = self._get_security_group_rule_object(context,
                                                                  rule)
        rulelist = groupd['logical_port_%s_rules' % direction]
        for check in checks:
            if not check(secrule, rulelist):
                raise checks[check]
        getattr(rulelist, operation)(secrule)

        LOG.debug("%s rule on security group %s" % (operation, groupd['uuid']))
        group = {'port_%s_rules' % direction: rulelist}
        return self.update_security_group(context, group_id, **group)

    def create_security_group_rule(self, context, group_id, rule):
        return self._update_security_group_rules(
            context, group_id, rule, 'append',
            {(lambda x, y: x not in y):
             sg_ext.SecurityGroupRuleExists(id=group_id),
             (lambda x, y:
                 self._check_rule_count_per_port(context, group_id) <
                 self.limits['max_rules_per_port']):
             q_exc.DriverLimitReached(limit="rules per port")})

    def delete_security_group_rule(self, context, group_id, rule):
        return self._update_security_group_rules(
            context, group_id, rule, 'remove',
            {(lambda x, y: x in y):
             sg_ext.SecurityGroupRuleNotFound(id="with group_id %s" %
                                              group_id)})

    def _create_or_choose_lswitch(self, context, network_id):
        switches = self._lswitch_status_query(context, network_id)
        switch = self._lswitch_select_open(context, network_id=network_id,
                                           switches=switches)
        if switch:
            LOG.debug("Found open switch %s" % switch)
            return switch

        switch_details = self._get_network_details(context, network_id,
                                                   switches)
        if not switch_details:
            raise q_exc.BadNVPState(net_id=network_id)

        return self._lswitch_create(context, network_id=network_id,
                                    **switch_details)

    def _lswitch_status_query(self, context, network_id):
        query = self._lswitches_for_network(context, network_id)
        query.relations("LogicalSwitchStatus")
        results = query.results()
        LOG.debug("Query results: %s" % results)
        return results

    def _lswitch_select_open(self, context, switches=None, **kwargs):
        """Selects an open lswitch for a network.

        Note that it does not select the most full switch, but merely one with
        ports available.
        """

        if switches is not None:
            for res in switches["results"]:
                count = res["_relations"]["LogicalSwitchStatus"]["lport_count"]
                if (self.limits['max_ports_per_switch'] == 0 or
                        count < self.limits['max_ports_per_switch']):
                    return res["uuid"]
        return None

    @utils.retry_loop(CONF.NVP.operation_retries)
    def _lswitch_delete(self, context, lswitch_uuid):
        with self.get_connection() as connection:
            LOG.debug("Deleting lswitch %s" % lswitch_uuid)
            connection.lswitch(lswitch_uuid).delete()

    def _config_provider_attrs(self, connection, switch, phys_net,
                               net_type, segment_id):
        if not (phys_net or net_type):
            return
        if not phys_net and net_type:
            raise q_exc.ProvidernetParamError(
                msg="provider:physical_network parameter required")
        if phys_net and not net_type:
            raise q_exc.ProvidernetParamError(
                msg="provider:network_type parameter required")
        if net_type not in ("bridge", "vlan", "vxlan") and segment_id:
            raise q_exc.SegmentIdUnsupported(net_type=net_type)
        if net_type == "vlan" and not segment_id:
            raise q_exc.SegmentIdRequired(net_type=net_type)

        phys_type = physical_net_type_map.get(net_type.lower())
        if not phys_type:
            raise q_exc.InvalidPhysicalNetworkType(net_type=net_type)

        tz_query = connection.transportzone(phys_net).query()
        transport_zone = tz_query.results()

        if transport_zone["result_count"] == 0:
            raise q_exc.PhysicalNetworkNotFound(phys_net=phys_net)
        switch.transport_zone(zone_uuid=phys_net,
                              transport_type=phys_type,
                              vlan_id=segment_id)

    def _add_default_tz_bindings(self, context, switch, network_id):
        """Configure any additional default transport zone bindings."""
        default_tz = CONF.NVP.default_tz

        # If there is no default tz specified it's pointless to try
        # and add any additional default tz bindings.
        if not default_tz:
            LOG.warn("additional_default_tz_types specified, "
                     "but no default_tz. Skipping "
                     "_add_default_tz_bindings().")
            return

        # This should never be called without a neutron network uuid,
        # we require it to bind some segment allocations.
        if not network_id:
            LOG.warn("neutron network_id not specified, skipping "
                     "_add_default_tz_bindings()")
            return

        for net_type in CONF.NVP.additional_default_tz_types:
            if net_type in TZ_BINDINGS:
                binding = TZ_BINDINGS[net_type]
                binding.add(context, switch, default_tz, network_id)
            else:
                LOG.warn("Unknown default tz type %s" % (net_type))

    def _remove_default_tz_bindings(self, context, network_id):
        """Deconfigure any additional default transport zone bindings."""
        default_tz = CONF.NVP.default_tz

        if not default_tz:
            LOG.warn("additional_default_tz_types specified, "
                     "but no default_tz. Skipping "
                     "_remove_default_tz_bindings().")
            return

        if not network_id:
            LOG.warn("neutron network_id not specified, skipping "
                     "_remove_default_tz_bindings()")
            return

        for net_type in CONF.NVP.additional_default_tz_types:
            if net_type in TZ_BINDINGS:
                binding = TZ_BINDINGS[net_type]
                binding.remove(context, default_tz, network_id)
            else:
                LOG.warn("Unknown default tz type %s" % (net_type))

    @utils.retry_loop(CONF.NVP.operation_retries)
    def _lswitch_create(self, context, network_name=None, tags=None,
                        network_id=None, phys_net=None,
                        phys_type=None, segment_id=None,
                        **kwargs):
        # NOTE(mdietz): physical net uuid maps to the transport zone uuid
        # physical net type maps to the transport/connector type
        # if type maps to 'bridge', then segment_id, which maps
        # to vlan_id, is conditionally provided
        LOG.debug("Creating new lswitch for %s network %s" %
                  (context.tenant_id, network_name))

        tenant_id = context.tenant_id
        with self.get_connection() as connection:
            switch = connection.lswitch()
            if network_name is None:
                network_name = network_id
            switch.display_name(network_name[:40])
            tags = tags or []
            tags.append({"tag": tenant_id, "scope": "os_tid"})
            if network_id:
                tags.append({"tag": network_id, "scope": "neutron_net_id"})
            switch.tags(tags)
            LOG.debug("Creating lswitch for network %s" % network_id)

            # TODO(morgabra) It seems like this whole interaction here is
            # broken. We force-add either the id/type from the network, or
            # the config default, *then* we still call _config_provider_attrs?
            # It seems like we should listen to the network then fall back
            # to the config defaults, but I'm leaving this as-is for now.
            pnet = phys_net or CONF.NVP.default_tz
            ptype = phys_type or CONF.NVP.default_tz_type
            switch.transport_zone(pnet, ptype)

            # When connecting to public or snet, we need switches that are
            # connected to their respective public/private transport zones
            # using a "bridge" connector. Public uses no VLAN, whereas private
            # uses VLAN 122 in netdev. Probably need this to be configurable
            self._config_provider_attrs(connection, switch, phys_net,
                                        phys_type, segment_id)

            # NOTE(morgabra) A hook for any statically-configured tz bindings
            # that should be added to the switch before create()
            # TODO(morgabra) I'm not sure the normal usage of provider net
            # attrs, and which should superscede which. This all needs a
            # refactor after some discovery probably.
            self._add_default_tz_bindings(context, switch, network_id)

            res = switch.create()
            try:
                uuid = res["uuid"]
                return uuid
            except TypeError:
                LOG.exception("Unexpected return from NVP: %s" % res)
                raise

    def get_lswitch_ids_for_network(self, context, network_id):
        """Public interface for fetching lswitch ids for a given network.

        NOTE(morgabra) This is here because calling private methods
        from outside the class feels wrong, and we need to be able to
        fetch lswitch ids for use in other drivers.
        """
        lswitches = self._lswitches_for_network(context, network_id).results()
        return [s['uuid'] for s in lswitches["results"]]

    def _lswitches_for_network(self, context, network_id):
        with self.get_connection() as connection:
            query = connection.lswitch().query()
            query.tagscopes(['os_tid', 'neutron_net_id'])
            query.tags([context.tenant_id, network_id])
            return query

    def _lswitch_from_port(self, context, port_id):
        with self.get_connection() as connection:
            query = connection.lswitch_port("*").query()
            query.relations("LogicalSwitchConfig")
            query.uuid(port_id)
            port = query.results()
            if port['result_count'] > 1:
                raise Exception("Could not identify lswitch for port %s" %
                                port_id)
            if port['result_count'] < 1:
                raise Exception("No lswitch found for port %s" % port_id)
            cfg = port['results'][0]["_relations"]["LogicalSwitchConfig"]
            return cfg["uuid"]

    def _get_security_group(self, context, group_id):
        with self.get_connection() as connection:
            query = connection.securityprofile().query()
            query.tagscopes(['os_tid', 'neutron_group_id'])
            query.tags([context.tenant_id, group_id])
            query = query.results()
            if query['result_count'] != 1:
                raise sg_ext.SecurityGroupNotFound(id=group_id)
            return query['results'][0]

    def _get_security_group_id(self, context, group_id):
        return self._get_security_group(context, group_id)['uuid']

    def _get_security_group_rule_object(self, context, rule):
        ethertype = rule.get('ethertype', None)
        rule_clone = {}

        ip_prefix = rule.get('remote_ip_prefix', None)
        if ip_prefix:
            rule_clone['ip_prefix'] = ip_prefix
        profile_uuid = rule.get('remote_group_id', None)
        if profile_uuid:
            rule_clone['profile_uuid'] = profile_uuid
        for key in ['protocol', 'port_range_min', 'port_range_max']:
            if rule.get(key):
                rule_clone[key] = rule[key]

        with self.get_connection() as connection:
            secrule = connection.securityrule(ethertype, **rule_clone)

            direction = rule.get('direction', '')
            if direction not in ['ingress', 'egress']:
                raise AttributeError(
                    "Direction not specified as 'ingress' or 'egress'.")
            return (direction, secrule)

    def _check_rule_count_per_port(self, context, group_id):
        with self.get_connection() as connection:
            ports = connection.lswitch_port("*").query().security_profile_uuid(
                '=', self._get_security_group_id(
                    context, group_id)).results().get('results', [])
            groups = (port.get('security_profiles', []) for port in ports)
            return max([self._check_rule_count_for_groups(
                context, (connection.securityprofile(gp).read()
                          for gp in group))
                        for group in groups] or [0])

    def _check_rule_count_for_groups(self, context, groups):
        return sum(len(group['logical_port_ingress_rules']) +
                   len(group['logical_port_egress_rules'])
                   for group in groups)

    def _get_security_groups_for_port(self, context, groups):
        if (self._check_rule_count_for_groups(
                context,
                (self._get_security_group(context, g) for g in groups))
                > self.limits['max_rules_per_port']):
            raise q_exc.DriverLimitReached(limit="rules per port")

        return [self._get_security_group(context, group)['uuid']
                for group in groups]



# Copyright (c) 2013 Rackspace Hosting Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from neutron import context
from neutron import wsgi
from oslo_log import log as logging
import webob.dec
import webob.exc

LOG = logging.getLogger(__name__)


class NoAuthMiddleware(wsgi.Middleware):
    """Return a fake token if one isn't specified."""

    @webob.dec.wsgify(RequestClass=wsgi.Request)
    def __call__(self, req):
        if 'X-Auth-Token' not in req.headers:
            user_id = req.headers.get('X-Auth-User', 'admin')
            project_id = req.headers.get('X-Auth-Project-Id', 'admin')
            res = webob.Response()
            res.headers['X-Auth-Token'] = '%s:%s' % (user_id, project_id)
            res.content_type = 'text/plain'
            res.status = '204'
            return res

        token = req.headers['X-Auth-Token']
        user_id, _sep, project_id = token.partition(':')
        project_id = project_id or user_id
        ctx = context.Context(user_id,
                              project_id,
                              is_admin=True)

        req.environ['neutron.context'] = ctx
        return self.application



"""Revise security group rules

Revision ID: 26e984b48a0d
Revises: 1664300cb03a
Create Date: 2014-09-16 22:01:07.329380

"""

# revision identifiers, used by Alembic.
revision = '26e984b48a0d'
down_revision = '1664300cb03a'

from alembic import op
import sqlalchemy as sa


OLD_TABLE = "quark_security_group_rule"
NEW_TABLE = "quark_security_group_rules"


def upgrade():
    # NOTE(mdietz): You can't change the datatype or remove columns,
    #               in SQLite, please see
    #               http://sqlite.org/lang_altertable.html
    op.drop_table(OLD_TABLE)
    op.create_table(
        NEW_TABLE,
        sa.Column('tenant_id', sa.String(length=255), nullable=True),
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('group_id', sa.String(length=36), nullable=False),
        sa.Column('direction', sa.String(length=10), nullable=False),
        sa.Column('port_range_max', sa.Integer(), nullable=True),
        sa.Column('port_range_min', sa.Integer(), nullable=True),
        sa.Column('protocol', sa.Integer(), nullable=True),
        sa.Column("ethertype", type_=sa.Integer(), nullable=False),
        sa.Column('remote_group_id', sa.String(length=36), nullable=True),
        sa.Column("remote_ip_prefix", type_=sa.String(255)),
        sa.ForeignKeyConstraint(["remote_group_id"],
                                ["quark_security_groups.id"],
                                "fk_remote_group_id"),
        sa.ForeignKeyConstraint(['group_id'], ['quark_security_groups.id'], ),
        sa.PrimaryKeyConstraint('id'),
        mysql_engine="InnoDB")


def downgrade():
    op.drop_table(NEW_TABLE)
    op.create_table(
        OLD_TABLE,
        sa.Column('tenant_id', sa.String(length=255), nullable=True),
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('group_id', sa.String(length=36), nullable=False),
        sa.Column('direction', sa.String(length=10), nullable=False),
        sa.Column('ethertype', sa.String(length=4), nullable=False),
        sa.Column('port_range_max', sa.Integer(), nullable=True),
        sa.Column('port_range_min', sa.Integer(), nullable=True),
        sa.Column('protocol', sa.Integer(), nullable=True),
        sa.Column('remote_ip_prefix', sa.String(length=22), nullable=True),
        sa.Column('remote_group_id', sa.String(length=36), nullable=True),
        sa.ForeignKeyConstraint(['group_id'], ['quark_security_groups.id'], ),
        sa.PrimaryKeyConstraint('id'),
        mysql_engine='InnoDB')



# Copyright 2013 Rackspace Hosting Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron.extensions import securitygroup as sg_ext
from neutron import quota
from neutron_lib import exceptions as n_exc
from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import uuidutils

from quark.db import api as db_api
from quark.environment import Capabilities
from quark import exceptions as q_exc
from quark import plugin_views as v
from quark import protocols

CONF = cfg.CONF
LOG = logging.getLogger(__name__)
DEFAULT_SG_UUID = "00000000-0000-0000-0000-000000000000"
GROUP_NAME_MAX_LENGTH = 255
GROUP_DESCRIPTION_MAX_LENGTH = 255


def _validate_security_group_rule(context, rule):
    # TODO(mdietz): As per RM8615, Remote groups are not currently supported
    if rule.get("remote_group_id"):
        raise n_exc.InvalidInput(
            error_message="Remote groups are not currently supported")

    direction = rule.get("direction")
    if direction == Capabilities.EGRESS:
        if Capabilities.EGRESS not in CONF.QUARK.environment_capabilities:
            raise q_exc.EgressSecurityGroupRulesNotEnabled()

    protocol = rule.pop('protocol')
    port_range_min = rule['port_range_min']
    port_range_max = rule['port_range_max']
    ethertype = protocols.translate_ethertype(rule["ethertype"])

    if protocol:
        protocol = protocols.translate_protocol(protocol, rule["ethertype"])
        protocols.validate_protocol_with_port_ranges(ethertype,
                                                     protocol,
                                                     port_range_min,
                                                     port_range_max)
        rule['protocol'] = protocol
    else:
        if port_range_min is not None or port_range_max is not None:
            raise sg_ext.SecurityGroupProtocolRequiredWithPorts()

    rule["ethertype"] = ethertype

    protocols.validate_remote_ip_prefix(ethertype,
                                        rule.get("remote_ip_prefix"))

    return rule


def _validate_security_group(security_group):
    if "name" in security_group:
        if len(security_group["name"]) > GROUP_NAME_MAX_LENGTH:
            raise n_exc.InvalidInput(
                error_message="Group name must be 255 characters or less")

        if security_group["name"] == "default":
            raise sg_ext.SecurityGroupDefaultAlreadyExists()

    if ("description" in security_group and
            len(security_group["description"]) > GROUP_DESCRIPTION_MAX_LENGTH):
        raise n_exc.InvalidInput(
            error_message="Group description must be 255 characters or less")


def create_security_group(context, security_group):
    LOG.info("create_security_group for tenant %s" %
             (context.tenant_id))
    group = security_group["security_group"]
    _validate_security_group(group)

    group_name = group.get('name', '')
    group_id = uuidutils.generate_uuid()

    with context.session.begin():
        group["id"] = group_id
        group["name"] = group_name
        group["tenant_id"] = context.tenant_id
        dbgroup = db_api.security_group_create(context, **group)
    return v._make_security_group_dict(dbgroup)


def create_security_group_rule(context, security_group_rule):
    LOG.info("create_security_group for tenant %s" %
             (context.tenant_id))
    with context.session.begin():
        rule = _validate_security_group_rule(
            context, security_group_rule["security_group_rule"])
        rule["id"] = uuidutils.generate_uuid()

        group_id = rule["security_group_id"]
        group = db_api.security_group_find(context, id=group_id,
                                           scope=db_api.ONE)
        if not group:
            raise sg_ext.SecurityGroupNotFound(id=group_id)

        quota.QUOTAS.limit_check(
            context, context.tenant_id,
            security_rules_per_group=len(group.get("rules", [])) + 1)

        new_rule = db_api.security_group_rule_create(context, **rule)
    return v._make_security_group_rule_dict(new_rule)


def delete_security_group(context, id):
    LOG.info("delete_security_group %s for tenant %s" %
             (id, context.tenant_id))

    with context.session.begin():
        group = db_api.security_group_find(context, id=id, scope=db_api.ONE)

        # TODO(anyone): name and ports are lazy-loaded. Could be good op later
        if not group:
            raise sg_ext.SecurityGroupNotFound(id=id)
        if id == DEFAULT_SG_UUID or group.name == "default":
            raise sg_ext.SecurityGroupCannotRemoveDefault()
        if group.ports:
            raise sg_ext.SecurityGroupInUse(id=id)
        db_api.security_group_delete(context, group)


def delete_security_group_rule(context, id):
    LOG.info("delete_security_group %s for tenant %s" %
             (id, context.tenant_id))
    with context.session.begin():
        rule = db_api.security_group_rule_find(context, id=id,
                                               scope=db_api.ONE)
        if not rule:
            raise sg_ext.SecurityGroupRuleNotFound(id=id)

        group = db_api.security_group_find(context, id=rule["group_id"],
                                           scope=db_api.ONE)
        if not group:
            raise sg_ext.SecurityGroupNotFound(id=id)

        rule["id"] = id
        db_api.security_group_rule_delete(context, rule)


def get_security_group(context, id, fields=None):
    LOG.info("get_security_group %s for tenant %s" %
             (id, context.tenant_id))
    group = db_api.security_group_find(context, id=id, scope=db_api.ONE)
    if not group:
        raise sg_ext.SecurityGroupNotFound(id=id)
    return v._make_security_group_dict(group, fields)


def get_security_group_rule(context, id, fields=None):
    LOG.info("get_security_group_rule %s for tenant %s" %
             (id, context.tenant_id))
    rule = db_api.security_group_rule_find(context, id=id,
                                           scope=db_api.ONE)
    if not rule:
        raise sg_ext.SecurityGroupRuleNotFound(id=id)
    return v._make_security_group_rule_dict(rule, fields)


def get_security_groups(context, filters=None, fields=None,
                        sorts=None, limit=None, marker=None,
                        page_reverse=False):
    LOG.info("get_security_groups for tenant %s" %
             (context.tenant_id))
    groups = db_api.security_group_find(context, **filters)
    return [v._make_security_group_dict(group) for group in groups]


def get_security_group_rules(context, filters=None, fields=None,
                             sorts=None, limit=None, marker=None,
                             page_reverse=False):
    LOG.info("get_security_group_rules for tenant %s" %
             (context.tenant_id))
    rules = db_api.security_group_rule_find(context, **filters)
    return [v._make_security_group_rule_dict(rule) for rule in rules]


def update_security_group(context, id, security_group):
    if id == DEFAULT_SG_UUID:
        raise sg_ext.SecurityGroupCannotUpdateDefault()
    new_group = security_group["security_group"]
    _validate_security_group(new_group)

    with context.session.begin():
        group = db_api.security_group_find(context, id=id, scope=db_api.ONE)
        db_group = db_api.security_group_update(context, group, **new_group)
    return v._make_security_group_dict(db_group)



# Copyright (c) 2013 Rackspace Hosting Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from neutron.api import extensions
from neutron import manager
from neutron import wsgi
from oslo_log import log as logging
import quark.utils as utils

RESOURCE_NAME = 'mac_address_range'
RESOURCE_COLLECTION = RESOURCE_NAME + "s"
EXTENDED_ATTRIBUTES_2_0 = {
    RESOURCE_COLLECTION: {}
}

attr_dict = EXTENDED_ATTRIBUTES_2_0[RESOURCE_COLLECTION]
attr_dict[RESOURCE_NAME] = {'allow_post': True,
                            'allow_put': False,
                            'is_visible': True}

LOG = logging.getLogger(__name__)


def mac_range_dict(mac_range):
    return dict(address=mac_range["cidr"],
                id=mac_range["id"])


class MacAddressRangesController(wsgi.Controller):

    def __init__(self, plugin):
        self._resource_name = RESOURCE_NAME
        self._plugin = plugin

    @utils.exc_wrapper
    def create(self, request, body=None):
        body = self._deserialize(request.body, request.get_content_type())
        return {"mac_address_range":
                self._plugin.create_mac_address_range(request.context, body)}

    @utils.exc_wrapper
    def index(self, request):
        context = request.context
        return {"mac_address_ranges":
                self._plugin.get_mac_address_ranges(context)}

    @utils.exc_wrapper
    def show(self, request, id):
        context = request.context
        return {"mac_address_range":
                self._plugin.get_mac_address_range(context, id)}

    @utils.exc_wrapper
    def delete(self, request, id, **kwargs):
        context = request.context
        return self._plugin.delete_mac_address_range(context, id)


class Mac_address_ranges(extensions.ExtensionDescriptor):
    """Mac Address Range support."""
    @classmethod
    def get_name(cls):
        return "MAC Address Ranges for a tenant"

    @classmethod
    def get_alias(cls):
        return RESOURCE_COLLECTION

    @classmethod
    def get_description(cls):
        return ("Expose functions for cloud admin to manage MAC ranges"
                "for each tenant")

    @classmethod
    def get_namespace(cls):
        return ("http://docs.openstack.org/network/ext/"
                "mac-address-ranges/api/v2.0")

    @classmethod
    def get_updated(cls):
        return "2013-02-19T10:00:00-00:00"

    def get_extended_resources(self, version):
        if version == "2.0":
            return EXTENDED_ATTRIBUTES_2_0
        else:
            return {}

    @classmethod
    def get_resources(cls):
        """Returns Ext Resources."""
        plugin = manager.NeutronManager.get_plugin()
        controller = MacAddressRangesController(plugin)
        return [extensions.ResourceExtension(Mac_address_ranges.get_alias(),
                                             controller)]



"""Implements reservations, resourcedeltas and quota usages from upstream

Revision ID: 2e9cf60b0ef6
Revises: 341d2e702dc4
Create Date: 2015-09-13 13:46:03.888079

"""

# revision identifiers, used by Alembic.
revision = '2e9cf60b0ef6'
down_revision = '341d2e702dc4'

from alembic import op
import sqlalchemy as sa
from sqlalchemy import sql


def upgrade():
    op.create_table(
        'reservations',
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('tenant_id', sa.String(length=255), nullable=True),
        sa.Column('expiration', sa.DateTime(), nullable=True),
        sa.PrimaryKeyConstraint('id'))

    op.create_table(
        'resourcedeltas',
        sa.Column('resource', sa.String(length=255), nullable=False),
        sa.Column('reservation_id', sa.String(length=36), nullable=False),
        sa.Column('amount', sa.Integer(), nullable=True),
        sa.ForeignKeyConstraint(['reservation_id'], ['reservations.id'],
                                ondelete='CASCADE'),
        sa.PrimaryKeyConstraint('resource', 'reservation_id'))

    op.create_table(
        'quotausages',
        sa.Column('tenant_id', sa.String(length=255),
                  nullable=False, primary_key=True, index=True),
        sa.Column('resource', sa.String(length=255),
                  nullable=False, primary_key=True, index=True),
        sa.Column('dirty', sa.Boolean(), nullable=False,
                  server_default=sql.false()),
        sa.Column('in_use', sa.Integer(), nullable=False,
                  server_default='0'),
        sa.Column('reserved', sa.Integer(), nullable=False,
                  server_default='0'))
